import time
from typing import Optional, AsyncGenerator, Union, List
import json
import logging
import asyncio
import uuid
from enum import Enum
from dotenv import load_dotenv
from pydantic import BaseModel
import httpx

from langfuse import observe
from langfuse.openai import AsyncOpenAI

from ..schemas import Message,ToolCall
from ..db import MessageDB
from ..core import Session
from ..utils.tool_decorator import function_tool
from ..utils.mcp_convertor import MCPTool

load_dotenv(override=True)

# Initialize logging(only needs to be done once)
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(name)s: %(message)s"
)

class ReplyType(Enum):
    SIMPLE_REPLY = "simple_reply"
    STRUCTURED_REPLY = "structured_reply"
    TOOL_CALL = "tool_call"
    ERROR = "error"

class Agent:
    """
    Base class for creating an AI agent that can interact with users, manage tools, and handle multi-step reasoning.
    """

    DEFAULT_NAME = "default_agent"
    DEFAULT_MODEL = "gpt-4.1-mini"

    DEFAULT_SYSTEM_PROMPT = (
        "**Context Information:**\n"
        "- Current user_id: {user_id}\n"
        "- Current date: {date}\n" 
        "- Current timezone: {timezone}\n\n"
        "**Basic Capabilities:**\n"
        "- Respond directly when you can answer a question without tools\n"
        "- Use available tools when specialized functionality is needed\n"
        "- Handle multi-step reasoning and break down complex problems\n"
        "- Be concise yet informative in your responses\n"
        "- When uncertain, ask clarifying questions\n"
        "- Acknowledge when a request is beyond your capabilities\n\n\n"
    )


    def __init__(
        self, 
        name: Optional[str] = None,
        system_prompt: Optional[str] = None,
        description: Optional[str] = None, # simple description of the agent for tool conversion , what the Agent does
        model: Optional[str] = None,
        client: Optional[AsyncOpenAI] = None,
        tools: Optional[list] = None,
        mcp_servers: Optional[Union[str, list]] = None,
        sub_agents: Optional[List[Union[tuple[str, str, str], 'Agent']]] = None,  # (name, description, server_url) or Agent instances
        output_type: type[BaseModel] = None
    ):
        """
        Initialize the Agent with optional parameters.
        Args:
            name (Optional[str]): The name of the agent.
            system_prompt (Optional[str]): Custom system prompt to prepend to the default.
            description (Optional[str]): Simple description of the agent for tool conversion.
            model (Optional[str]): The OpenAI model to use, defaults to DEFAULT_MODEL.
            client (Optional[AsyncOpenAI]): Custom OpenAI client instance, defaults to a new AsyncOpenAI instance.
            tools (Optional[list]): List of tool functions to register with the agent.
            mcp_servers (Optional[Union[str, list]]): MCP server URLs to fetch tools from.
            sub_agents (Optional[List[Union[tuple[str, str, str], 'Agent']]]): List of sub-agents to convert to tools.
            output_type (type[BaseModel]): Pydantic model for structured output, can be overridden in chat method.

        Initializes the agent with a name, system prompt, description, model, and tools.
        If no name is provided, it defaults to "default_agent". The system prompt is a combination of a default prompt and any custom prompt provided.
        The agent can also register tools, fetch tools from MCP servers, and convert sub-agents into tools.
        """
        self.name: str = name or self.DEFAULT_NAME
        self.system_prompt: str = self.DEFAULT_SYSTEM_PROMPT + (system_prompt or "")
        self.description: str = description
        self.model: str = model or self.DEFAULT_MODEL
        self.client: AsyncOpenAI = client or AsyncOpenAI()
        self.tools: dict = {}
        self._register_tools(tools or [])
        self.mcp_servers: list = mcp_servers or []
        self.mcp_tools: dict = {}
        self.mcp_tools_last_updated: Optional[float] = None
        self.mcp_cache_ttl: int = 300  # 5 minutes
        self.agent_tools = self._convert_sub_agents_to_tools(sub_agents)
        self._register_tools(self.agent_tools)
        self.output_type: type[BaseModel] = output_type
        self.logger = logging.getLogger(self.__class__.__name__)
        if self.agent_tools:
            self.logger.info("Registered agent tools: %s", [tool.tool_spec['name'] for tool in self.agent_tools])

    async def __call__(
            self,
            user_message: Message | str,
            session: Session,
            history_count: int = 16,
            max_iter: int = 5,
            image_source: Optional[str] = None,
            output_type: type[BaseModel] = None,
            stream=False
    ) -> str | BaseModel | AsyncGenerator[str, None]:
        """
        Call the agent to generate a reply based on the user message and session.
        """
        return await self.chat(
            user_message=user_message,
            session=session,
            history_count=history_count,
            max_iter=max_iter,
            image_source=image_source,
            output_type=output_type,
            stream=stream
        )

    @observe()
    async def chat(
        self,
        user_message: Message | str,
        session: Session,
        history_count: int = 16,
        max_iter: int = 10,
        image_source: Optional[str] = None,
        output_type: type[BaseModel] = None,
        stream=False
    ) -> str | BaseModel | AsyncGenerator[str, None]:
        """
        Generate a reply from the agent given a user message and session.

        Args:
            user_message (Message | str): The latest user message.
            session (Session): The session object managing message history.
            history_count (int, optional): Number of previous messages to include. Defaults to 20.
            max_iter (int, optional): Maximum model call attempts. Defaults to 10.
            image_source (Optional[str], optional): Source of the image, if any can be a URL or File path or base64 string.
            output_type (type[BaseModel], optional): Pydantic model for structured output. Defaults to the Agent setting but can be overridden in chat().
            stream (bool, optional): Whether to stream the response. Defaults to False.

        Returns:
            str | BaseModel | AsyncGenerator[str, None]: The agent's reply or error message.
        """

        if output_type is None:
            output_type = self.output_type
            
        if output_type:
            stream = False  # Structured output does not support streaming

        try:
            # Register tools and MCP servers in each chat call to make sure they are up-to-date
            await self._register_mcp_servers(self.mcp_servers)

            # Store the incoming user message in session history
            await self._store_user_message(user_message, session, image_source)

            # Build input messages once outside the loop
            input_messages = [msg.to_dict() for msg in await session.get_messages(history_count)]

            for attempt in range(max_iter):

                reply_type, response = await self._call_model(input_messages, session, output_type, stream)

                if reply_type == ReplyType.SIMPLE_REPLY:
                    if not stream:
                        await self._store_model_reply(str(response), session)
                    return response
                elif reply_type == ReplyType.STRUCTURED_REPLY:
                    await self._store_model_reply(response.model_dump_json(), session)
                    return response
                elif reply_type == ReplyType.TOOL_CALL:
                    await self._handle_tool_calls(response, session, input_messages)
                else:
                    self.logger.error("Unknown reply type: %s", reply_type)
                    return "Sorry, I encountered an error while processing your request."

            # If no valid reply after max_iter attempts
            self.logger.error("Failed to generate response after %d attempts", max_iter)
            return "Sorry, I could not generate a response after multiple attempts."

        except Exception as e:
            self.logger.exception("Agent chat error: %s", e)
            return "Sorry, something went wrong."

    def as_tool(self,name: str = None, description: str = None,message_db: Optional[MessageDB] = None):
        """
        Convert the agent into an OpenAI tool function.
        Args:
            name (str): The name of the tool function.
            description (str): A brief description of what the tool does.
            message_db (Optional[MessageDB]): Optional message database for storing messages.
        Returns:
            function: An asynchronous function that can be used as an OpenAI tool.
        """
        @function_tool(name=name or self.name, description=description or self.description)
        async def tool_func(simple_input: str, shared_context: Optional[str] = None, image_source: Optional[str] = None):
            return await self.chat(user_message=f"### Shared Context:\n{shared_context}\n\n### User Input:\n{simple_input}" if shared_context else simple_input,
                                   image_source=image_source,
                                   session=Session(user_id=f"agent_{self.name}_as_tool", 
                                                   session_id=f"{uuid.uuid4()}",
                                                    message_db=message_db
                                                   ))
        return tool_func

    def _convert_sub_agents_to_tools(self, sub_agents: Optional[List[Union[tuple[str, str, str], 'Agent']]]) -> Optional[list]:
        """
        Convert sub-agents into OpenAI tool functions.
        Args:
            sub_agents (Optional[List[Union[tuple[str, str, str], 'Agent']]]): List of sub-agents to convert.
            Each item can be a tuple (name, description, server_url) or an Agent instance.
        Returns:
            Optional[list]: List of tool functions converted from sub-agents.
        """
        tools = []
        for item in sub_agents or []:
            if isinstance(item, tuple) and len(item) == 3:
                name, description, server = item
                tool = self._convert_http_agent_to_tool(server=server, name=name, description=description)
                tools.append(tool)
            elif isinstance(item, Agent):
                tool = item.as_tool()
                tools.append(tool)
            else:
                self.logger.warning(f"Invalid sub_agent type: {type(item)}. Must be tuple[name, description, server_url] or Agent instance.")
        return tools if tools else None

    def _register_tools(self, tools: Optional[list]) -> None:
        """
        Register tool functions with the agent.
        Ensures each tool function is asynchronous and unique.
        Args:
            tools (Optional[list]): List of tool functions to register.
        """
        for fn in tools or []:
            if not asyncio.iscoroutinefunction(fn):
                raise TypeError(f"Tool function '{fn.tool_spec['name']}' must be async.")
            if fn.tool_spec['name'] not in self.tools:
                self.tools[fn.tool_spec['name']] = fn

    @observe()
    async def _register_mcp_servers(self, mcp_servers: Optional[Union[str, list]]) -> None:
        """
        Register tools from MCP servers, updating the local cache if needed.
        Args:
            mcp_servers (Optional[Union[str, list]]): MCP server URLs to fetch tools from.
        """
        now = time.time()
        if self.mcp_tools_last_updated and (now - self.mcp_tools_last_updated) < self.mcp_cache_ttl:
            return

        self.mcp_tools = {}
        if isinstance(mcp_servers, str):
            mcp_servers = [mcp_servers]
        for url in mcp_servers or []:
            try:
                mt = MCPTool(url)
                mcp_tools = await mt.get_openai_tools()
                for tool in mcp_tools:
                    if tool.tool_spec['name'] not in self.mcp_tools:
                        self.mcp_tools[tool.tool_spec['name']] = tool
            except Exception as e:
                self.logger.error(f"Failed to get tools from MCP server {url}: {e}")
                # If one server fails, we should probably not update the timestamp
                # to try again on the next call. But for now, we continue with other servers.
                continue
        
        self.mcp_tools_last_updated = now

    async def _store_user_message(self, user_message: Message | str, session: Session, image_source: Optional[str]) -> None:
        if isinstance(user_message, str):
            user_message = Message.create(content=user_message, role="user", image_source=image_source)
        await session.add_messages(user_message)

    async def _store_model_reply(self, reply_text: str, session: Session) -> None:
        model_msg = Message.create(content=reply_text, role="assistant")
        await session.add_messages(model_msg)

    @observe()
    async def _call_model(self, input_msgs: list, session: Session, 
                          output_type: type[BaseModel] = None, stream: bool = False) -> tuple[ReplyType, object]:
        """
        Call the AI model with the provided input messages and session.
        Args:
            input_msgs (list): List of input messages to send to the model.
            session (Session): The session object managing message history.
            output_type (type[BaseModel], optional): Pydantic model for structured output.
            stream (bool, optional): Whether to stream the response. Defaults to False.
        Returns:
            tuple[ReplyType, object]: A tuple containing the reply type and the response object.
        """
        system_msg = {
            "role": "system",
            "content": self.system_prompt.format(
                user_id=session.user_id, 
                date=time.strftime('%Y-%m-%d'), 
                timezone=time.tzname[0]
            )
        }

        # 预先构建工具规格列表，避免重复计算
        all_tools = list(self.tools.values()) + list(self.mcp_tools.values())
        tool_specs = [fn.tool_spec for fn in all_tools] if all_tools else None
        
        # 预处理消息
        messages = [system_msg] + self._sanitize_input_messages(input_msgs)

        try:
            # 根据是否需要结构化输出选择不同的API调用,结构化输出强制不使用Stream 模式
            if output_type is not None:
                response = await self.client.responses.parse(
                    model=self.model,
                    tools=tool_specs,
                    input=messages,
                    text_format=output_type
                )
                # 检查结构化输出
                if hasattr(response, "output_parsed") and response.output_parsed is not None:
                    return ReplyType.STRUCTURED_REPLY, response.output_parsed
            else:
                response = await self.client.responses.create(
                    model=self.model,
                    tools=tool_specs,
                    input=messages,
                    stream=stream
                )

            # 统一处理响应，按优先级检查不同类型的输出
            if not stream:
                if hasattr(response, 'output_text') and response.output_text:
                    return ReplyType.SIMPLE_REPLY, response.output_text
                
                if hasattr(response, 'output') and response.output:
                    return ReplyType.TOOL_CALL, response.output
                
                # 如果没有有效输出，记录警告并返回错误
                self.logger.warning("Model response contains no valid output: %s", response)
                return ReplyType.ERROR, "No valid output from model response."
            else:
                
                # Get the third event to determine the stream type
                await anext(response, None)  # Skip first event
                await anext(response, None)  # Skip second event
                third_event = await anext(response, None)
                event_type = third_event.item.type if third_event else None

                if event_type == "message":
                    async def stream_generator():
                        async for event in response:
                            if event.type == 'response.output_text.delta':
                                content = event.delta
                                if content:
                                    yield content
                        await self._store_model_reply(event.response.output[0].content[0].text, session)
                    return ReplyType.SIMPLE_REPLY, stream_generator()
                elif event_type == "function_call":
                    async for event in response:
                        pass 
                    return ReplyType.TOOL_CALL, event.response.output
                else:
                    self.logger.warning("Stream response event_type is not recognized: %s", event_type)
                    async def stream_generator():
                        yield "Stream response event_type is not recognized."
                    # 返回一个生成器，避免直接返回错误信息
                    return ReplyType.ERROR, stream_generator()
                    
        except Exception as e:
            self.logger.exception("Model call failed: %s", e)
            if stream:
                async def stream_generator():
                    yield f"Model call error: {str(e)}"
                return ReplyType.ERROR, stream_generator()
            return ReplyType.ERROR, f"Model call error: {str(e)}"
    
    @observe()
    async def _handle_tool_calls(self, tool_calls: list, session: Session, input_messages: list) -> None:
        """
        Handle tool calls by executing them concurrently.
        Args:
            tool_calls (list): List of tool call messages to process.
            session (Session): The session object managing message history.
            input_messages (list): List of input messages to update with tool call results.
        Returns:
            None: This method modifies input_messages in place and does not return a value.
        """

        if tool_calls is None or not tool_calls:
            return None

        tasks = [self._act(tc, session) for tc in tool_calls if getattr(tc, "type", None) == "function_call"]
        results = await asyncio.gather(*tasks)
        # Safely add all tool messages after concurrent execution
        for tool_messages in results:
            if tool_messages:
                input_messages.extend([msg.to_dict() for msg in tool_messages])
        return None

    async def _act(self, tool_call, session: Session) -> Optional[list]:
        """
        Execute a single tool call and return the messages generated by the tool.
        Args:
            tool_call: The tool call message to process.
            session (Session): The session object managing message history.
        Returns:
            Optional[list]: A list of messages generated by the tool call, or None if the tool is not found or an error occurs.
        """
        name = getattr(tool_call, "name", None)
        try:
            args = json.loads(getattr(tool_call, "arguments", "{}"))
        except Exception as e:
            self.logger.error("Tool args parse error: %s", e)
            return None
        func = self.tools.get(name) or self.mcp_tools.get(name)
        if func:
            self.logger.info("Calling tool: %s with args: %s", name, args)

            try:
                result = await func(**args)
            except Exception as e:
                self.logger.error("Tool call error: %s", e)
                result = f"Tool error: {e}"

            tool_call_msg = Message(
                type="function_call",
                role="tool", 
                content=f"Calling tool: `{name}` with args: {args}",
                tool_call=ToolCall(
                    call_id=getattr(tool_call, "call_id", ""),
                    name=name,
                    arguments=json.dumps(args, ensure_ascii=False)
                )
            )

            tool_res_msg = Message(
                type="function_call_output",
                role="tool",
                content=f"Tool `{name}` result: {str(result) if len(str(result)) < 20 else str(result)[:20] + '...'}",
                tool_call=ToolCall(
                    call_id=getattr(tool_call, "call_id", "001"),
                    output=json.dumps(result, ensure_ascii=False) if isinstance(result, (dict, list)) else str(result)
                )
            )
            await session.add_messages([tool_call_msg, tool_res_msg])
            
            # Return the messages instead of modifying input_messages directly
            return [tool_call_msg, tool_res_msg]

        return None

    @staticmethod
    def _sanitize_input_messages(input_messages: list) -> list:
        """
        Checks if the first message is a function call output and removes it if so.
        Args:
            input_messages (list): List of input messages to sanitize.
        Returns:
            list: Sanitized list of input messages with no leading function call outputs.
        """
        while input_messages and input_messages[0].get("type") == "function_call_output":
            input_messages.pop(0)
        return input_messages
    

    def _convert_http_agent_to_tool(self, server: str, name: str = None, description: str = None):
        """
        Convert an HTTP-based agent into an OpenAI tool function.
        Args:
            server (str): The URL of the HTTP agent server.
            name (str): The name of the tool function.
            description (str): A brief description of what the tool does.
        Returns:
            function: An asynchronous function that can be used as an OpenAI tool.
        """
        @function_tool(name=name, description=description)
        async def tool_func(simple_input: str, shared_context: Optional[str] = None, image_source: Optional[str] = None):
            request_body = {
                "user_id": f"http_tool_{uuid.uuid4().hex[:8]}",
                "session_id": f"session_{uuid.uuid4().hex[:8]}",
                "user_message": f"### Shared Context:\n{shared_context}\n\n### User Input:\n{simple_input}" if shared_context else simple_input,
                "stream": False  # 工具调用不使用流式响应
            }
            
            if image_source:
                request_body["image_source"] = image_source
            
            chat_url = f"{server}/chat"
            
            self.logger.info(f"Calling HTTP Agent at: {chat_url}")
            self.logger.debug(f"Request body: {request_body}")

            try:
                async with httpx.AsyncClient(timeout=60.0) as client:  # 增加超时时间
                    response = await client.post(chat_url, json=request_body)
                    
                    self.logger.debug(f"HTTP response status: {response.status_code}")
                    
                    if response.status_code == 200:
                        try:
                            response_data = response.json()
                            reply = response_data.get("reply", "")
                            
                            if reply:
                                self.logger.info(f"HTTP Agent '{name}' responded successfully")
                                return reply
                            else:
                                error_msg = "Empty reply from HTTP Agent"
                                self.logger.warning(error_msg)
                                return error_msg
                                
                        except json.JSONDecodeError as e:
                            error_msg = f"Invalid JSON response from HTTP Agent: {str(e)}"
                            self.logger.error(f"{error_msg}. Response text: {response.text[:200]}...")
                            return error_msg
                            
                    elif response.status_code == 500:
                        try:
                            error_data = response.json()
                            error_detail = error_data.get("detail", "Internal server error")
                            error_msg = f"HTTP Agent internal error: {error_detail}"
                        except:
                            error_msg = f"HTTP Agent internal error: {response.text[:200]}"
                        
                        self.logger.error(error_msg)
                        return error_msg
                        
                    else:
                        error_msg = f"HTTP Agent error {response.status_code}: {response.text[:200]}"
                        self.logger.error(error_msg)
                        return error_msg
                        
            except httpx.ConnectError as e:
                error_msg = f"Cannot connect to HTTP Agent at {chat_url}: {str(e)}"
                self.logger.error(error_msg)
                return error_msg
                
            except httpx.TimeoutException as e:
                error_msg = f"HTTP Agent request timeout: {str(e)}"
                self.logger.error(error_msg)
                return error_msg
                
            except httpx.RequestError as e:
                error_msg = f"HTTP Agent request error: {str(e)}"
                self.logger.error(error_msg)
                return error_msg
                
            except Exception as e:
                error_msg = f"Unexpected error calling HTTP Agent: {str(e)}"
                self.logger.exception(error_msg)
                return error_msg
        
        return tool_func


import time
from pydantic import BaseModel, Field
from typing import Optional

from ..utils.image_upload import upload_image

class ToolCall(BaseModel):
    """Represents a tool/function call within a message."""
    call_id: str = Field(..., description="Call ID for tracking")
    name: Optional[str] = Field(None, description="Name of the function/tool being called")
    arguments: Optional[str] = Field(None, description="Arguments for the function call, as a JSON string")
    output: Optional[str] = Field(None, description="Output/result of the function call")

class ImageContent(BaseModel):
    """Represents image content in a message."""
    format: str = Field(..., description="Image format (e.g., png, jpeg)")
    source: Optional[str] = Field(None, description="URL or base64 string of the image")

class VoiceContent(BaseModel):
    """Represents voice content in a message."""
    format: str = Field(..., description="Voice format (e.g., mp3, wav)")
    source: Optional[bytes] = Field(None, description="The binary content of the voice file")

class DocumentContent(BaseModel):
    """Represents document content in a message."""
    format: str = Field(..., description="Document format (e.g., pdf, docx)")
    source: Optional[bytes] = Field(None, description="The binary content of the document")

class MultiModalContent(BaseModel):
    """Represents multi-modal content in a message."""
    image: Optional[ImageContent] = Field(None, description="Image content associated with the message")
    voice: Optional[VoiceContent] = Field(None, description="Voice content associated with the message")
    document: Optional[DocumentContent] = Field(None, description="Document content associated with the message")

class Message(BaseModel):
    """Message model for communication between roles."""
    type: str = Field("message", description="Type of message (e.g., message, function_call)")
    role: str = Field(..., description="The role of the sender (e.g., user, assistant)")
    content: str = Field(..., description="The content of the message")
    timestamp: float = Field(default_factory=time.time, description="The timestamp of when the message was sent")
    tool_call: Optional[ToolCall] = Field(None, description="tool/function calls associated with the message")
    multimodal: Optional[MultiModalContent] = Field(None, description="Multi-modal content associated with the message")

    @classmethod
    def create(
        cls,
        content: str,
        role: Optional[str] = "user",
        image_source: Optional[str] = None,
    ) -> "Message":
        """
        Create a message with optional image content.
        Args:
            content (str): The text content of the message.
            role (Optional[str]): The role of the sender (default is "user").
            image_source (Optional[str]): The URL or file path or base64 string of the image to be included in the message.
        Returns:
            Message: An instance of the Message class with the provided content and optional image.

        Raises:
            ValueError: If both image_url and image_source are provided.

        Usage:
            # Create a text message
            msg = Message.create("Hello, world!")
            # Create a message with specific role
            msg = Message.create("Hello, world!", role="assistant")
            # Create a message with an image URL
            msg = Message.create("Hello, world!", image_source="https://example.com/image.jpg")
        """
        multimodal = None
        if image_source:
            if not (image_source.startswith("http") or image_source.startswith("data:image/")):
                uploaded_url = upload_image(image_source)
                if uploaded_url:
                    image_source = uploaded_url
                else:
                    raise ValueError("Image upload failed, please check the image source.")
            multimodal = MultiModalContent(
                image=ImageContent(format="jpeg", source=image_source)
            )

        return cls(
            role=role,
            type="message",
            content=content,
            multimodal=multimodal
        )

    def to_dict(self) -> dict:
        """Convert the message to a dictionary, including tool call if present."""
        if self.type == "message":
            if self.multimodal and self.multimodal.image:
                return {
                    "role": self.role,
                    "content": [
                        {"type": "input_text", "text": self.content},
                        {
                            "type": "input_image",
                            "image_url": self.multimodal.image.source,
                        },
                    ],
                }
            return {
                "role": self.role,
                "content": self.content,
            }
        elif self.type in ["function_call", "function_call_output"]:
            result = {
            "call_id": self.tool_call.call_id,
            "type": self.type,
            "name": self.tool_call.name,
            "arguments": self.tool_call.arguments,
            "output": self.tool_call.output
            }
            # Filter out keys with value None
            return {k: v for k, v in result.items() if v is not None}
        else:
            raise ValueError(f"Unsupported message type: {self.type}")


#!/usr/bin/env python3
"""
Launcher script for xAgent Streamlit web interface.
This script properly launches the Streamlit app using streamlit run.
"""

import sys
import subprocess
import os
import argparse
from pathlib import Path

def main():
    """Launch the Streamlit app using streamlit run command."""
    # Parse command line arguments
    parser = argparse.ArgumentParser(description="Launch xAgent Web UI")
    parser.add_argument(
        "--agent-server",
        default="http://localhost:8010",
        help="URL of the xAgent server (default: http://localhost:8010)"
    )
    parser.add_argument(
        "--host",
        default="0.0.0.0",
        help="Host address for Streamlit server (default: 0.0.0.0)"
    )
    parser.add_argument(
        "--port",
        default="8501",
        help="Port for Streamlit server (default: 8501)"
    )
    
    args = parser.parse_args()
    # Get the directory where this launcher script is located
    script_dir = Path(__file__).parent
    app_path = script_dir / "app.py"
    
    # Ensure the app.py file exists
    if not app_path.exists():
        print(f"Error: Could not find app.py at {app_path}")
        sys.exit(1)
    
    # Set environment variables for the Streamlit app
    env = os.environ.copy()
    env["XAGENT_SERVER_URL"] = args.agent_server
    
    # Build the streamlit run command
    cmd = [
        sys.executable, "-m", "streamlit", "run", 
        str(app_path),
        "--server.headless", "true",
        "--server.port", args.port,
        "--server.address", args.host
    ]
    
    try:
        # Launch streamlit
        print("🚀 Starting xAgent Web UI...")
        print(f"📍 The web interface will be available at: http://{args.host}:{args.port}")
        print(f"🔗 Agent server URL: {args.agent_server}")
        print("🔧 Use Ctrl+C to stop the server")
        print()
        subprocess.run(cmd, check=True, env=env)
    except subprocess.CalledProcessError as e:
        print(f"Error launching Streamlit app: {e}")
        sys.exit(1)
    except KeyboardInterrupt:
        print("\nShutting down xAgent Web UI...")
        sys.exit(0)

if __name__ == "__main__":
    main()



import logging
from typing import List, Optional

from ..schemas import Message
from ..db import MessageDB

class Session:
    """
    Session class to manage user sessions and message history.
    """
    _local_messages = {}  # {(user_id, session_id): [Message, ...]}

    def __init__(
        self,
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        message_db: Optional[MessageDB] = None
    ):
        self.user_id = user_id or "default_user"
        self.session_id = session_id or "default_session"
        self.message_db = message_db
        key = (self.user_id, self.session_id)
        if not self.message_db and key not in Session._local_messages:
            Session._local_messages[key] = []
        self.logger = logging.getLogger(f"{self.__class__.__name__}[{self.user_id}:{self.session_id}]")

    async def add_messages(self, messages: Message | List[Message]) -> None:
        """
        Add messages to the session history.
        Args:
            messages: A single Message object or a list of Message objects to add.
        """
        try:
            if not isinstance(messages, list):
                messages = [messages]
            if self.message_db:
                self.logger.info("Adding messages to DB: %s", messages)
                await self.message_db.add_messages(self.user_id, messages, self.session_id)
            else:
                key = (self.user_id, self.session_id)
                self.logger.info("Adding messages to local session: %s", messages)
                max_local_history = 100
                Session._local_messages[key].extend(messages)
                if len(Session._local_messages[key]) > max_local_history:
                    Session._local_messages[key] = Session._local_messages[key][-max_local_history:]
        except Exception as e:
            self.logger.error("Failed to add messages: %s", e)

    async def get_messages(self, count: int = 20) -> List[Message]:
        """
        Get the last `count` messages from the session history.
        Args:
            count: Number of messages to retrieve (default is 20).
        Returns:
            List[Message]: List of Message objects from the session history.
        """
        try:
            if self.message_db:
                self.logger.info("Fetching last %d messages from DB", count)
                return await self.message_db.get_messages(self.user_id, self.session_id, count)
            key = (self.user_id, self.session_id)
            self.logger.info("Fetching last %d messages from local session", count)
            return Session._local_messages[key][-count:]
        except Exception as e:
            self.logger.error("Failed to get history: %s", e)
            return []

    async def clear_session(self) -> None:
        """
        Clear the session history.
        This will remove all messages from the session.
        """
        try:
            if self.message_db:
                self.logger.info("Clearing history in DB")
                await self.message_db.clear_history(self.user_id, self.session_id)
            else:
                key = (self.user_id, self.session_id)
                self.logger.info("Clearing local session history")
                Session._local_messages[key] = []
        except Exception as e:
            self.logger.error("Failed to clear history: %s", e)

    async def pop_message(self) -> Optional[Message]:
        """
        Pop the last message from the session history.
        This will remove the last message from the session and return it.
        If the last message is a tool result, it will continue popping until a non-tool result message is found.
        If no such message exists, it returns None.
        Returns:
            Optional[Message]: The last non-tool result message or None if no such message exists.
        """
        try:
            if self.message_db:
                self.logger.info("Popping last message from DB")
                return await self.message_db.pop_message(self.user_id, self.session_id)
            else:
                key = (self.user_id, self.session_id)
                self.logger.info("Popping last message from local session")
                while Session._local_messages[key]:
                    msg = Session._local_messages[key].pop()
                    if not msg.tool_call:
                        return msg
                return None
        except Exception as e:
            self.logger.error("Failed to pop message: %s", e)
            return None



"""
xAgent - Multi-Modal AI Agent System

A powerful multi-modal AI Agent system with modern architecture.
"""

from .core import Session, Agent, HTTPAgentServer
from .schemas import Message, ToolCall
from .db import MessageDB
from .utils import function_tool
from .tools import web_search, draw_image
from .multi import Swarm, Workflow
from .__version__ import __version__

__all__ = [
    # Core components
    "Session",
    "Agent", 
    "HTTPAgentServer",
    
    # Data models
    "Message",
    "ToolCall",
    "MessageDB",
    
    # Utilities
    "function_tool",

    # Built-in tools
    "web_search",
    "draw_image",

    # Multi-agent
    "Swarm",
    "Workflow",
    
    # Meta
    "__version__"
]



from pydantic import BaseModel, Field
from typing import Optional, Dict, List
from enum import Enum

class DifficultyLevel(str, Enum):
    """Enumeration for vocabulary difficulty levels."""
    BEGINNER = "beginner"        # Easy (A1-A2)
    INTERMEDIATE = "intermediate" # Intermediate (B1-B2)
    ADVANCED = "advanced"        # Advanced (C1-C2)
    EXPERT = "expert"            # Expert

class BaseVocabularyRecord(BaseModel):
    """Vocabulary record base model."""
    word: str = Field(..., description="The vocabulary word in lowercase")
    explanation: str = Field(..., description="The explanation of the word")
    example_sentences: List[str] = Field(default_factory=list, description="The example sentences")
    difficulty_level: DifficultyLevel = Field(default=DifficultyLevel.INTERMEDIATE, description="The difficulty level of the word")

class VocabularyRecord(BaseVocabularyRecord):
    """Vocabulary record model."""
    user_id: str = Field(..., description="The user ID")
    create_timestamp: Optional[float] = Field(default=None, description="The creation timestamp")
    update_timestamp: Optional[float] = Field(default=None, description="The last update timestamp")
    last_reviewed_timestamp: Optional[float] = Field(default=None, description="The last reviewed timestamp")
    familiarity: int = Field(0, ge=0, le=10, description="The familiarity level of the word, from 0 to 10")
    extra: Optional[Dict[str, str]] = Field(default=None, description="The extra fields for storing additional information")



from collections import Counter
from xagent.utils.tool_decorator import function_tool

@function_tool()
def char_count(text: str) -> dict:
    """
    Count the frequency of each character in the given text.
    
    Args:
        text (str): The input text to analyze.
        
    Returns:
        dict: A dictionary with characters as keys and their counts as values.
    """
    if not text:
        return {}
    
    # Use Counter to count character frequencies
    return dict(Counter(text))


from fastmcp import Client
from typing import Any, Callable, List

class MCPTool:
    """
    MCPTool 对象，初始化时传入 fastmcp 服务的 URL，可获取 openai 规范的 tools。
    """

    def __init__(self, url: str):
        """
        Args:
            url: fastmcp 服务的 URL
        """
        self.client = Client(url)

    def _get_attr_or_key(self, obj, key, default=None):
        # Try attribute first, then dict key, else default
        if hasattr(obj, key):
            return getattr(obj, key)
        elif isinstance(obj, dict):
            return obj.get(key, default)
        else:
            return default

    def _convert_to_openai_tool(self, mcp_tool: Any) -> Callable:
        """
        将 MCP tool 对象或 dict 转换为 OpenAI tool 规范的异步函数，并带有 tool_spec 属性。
        """
        name = self._get_attr_or_key(mcp_tool, "name")
        description = self._get_attr_or_key(mcp_tool, "description", "")
        input_schema = self._get_attr_or_key(mcp_tool, "inputSchema", {})
        input_schema.update({"additionalProperties": False})

        async def openai_tool_func(**kwargs):
            if self.client is None:
                raise RuntimeError("MCPTool: client is not set.")
            async with self.client:
                result = await self.client.call_tool(name, kwargs)
                return result.data

        openai_tool_func.tool_spec = {
            "type": "function",
            "name": name,
            "description": description,
            "parameters": input_schema
        }
        return openai_tool_func

    async def get_openai_tools(self) -> List[Callable]:
        """
        获取 openai 规范的 tool 函数列表（带 tool_spec 属性）。
        Returns:
            List[Callable]: 每个函数带有 tool_spec 属性
        """
        async with self.client:
            tools = await self.client.list_tools()
            return [self._convert_to_openai_tool(tool) for tool in tools]

if __name__ == "__main__":
    import asyncio

    mcp_tool = MCPTool("http://127.0.0.1:8001/mcp/")

    async def main():
        tools = await mcp_tool.get_openai_tools()
        for tool in tools:
            print(f"Tool: {tool.tool_spec['name']}, Description: {tool.tool_spec['description']}")
        result = await tools[0](n_dice=2)
        print(f"Result: {result}")

    asyncio.run(main())


"""
Frontend components for xAgent.

This module provides web interfaces for interacting with xAgent:
- Streamlit chat application for conversational interface
- Configuration UI for visual agent setup and management
"""

from .app import main as chat_app_main
from .launcher import main as web_launcher_main
from .config_ui import main as config_ui_main
from .config_launcher import main as config_launcher_main

__all__ = [
    "chat_app_main",
    "web_launcher_main", 
    "config_ui_main",
    "config_launcher_main"
]



#!/usr/bin/env python3
"""
xAgent Configuration UI Launcher

A Streamlit-based web interface for configuring and managing xAgent HTTP servers.
"""

import sys
import argparse
import subprocess
import socket
from pathlib import Path


def is_port_available(host, port):
    """Check if a port is available for binding."""
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.bind((host, port))
            return True
    except OSError:
        return False


def find_available_port(host, start_port, max_attempts=10):
    """Find an available port starting from start_port."""
    for i in range(max_attempts):
        port = start_port + i
        if is_port_available(host, port):
            return port
    raise RuntimeError(f"Could not find an available port after {max_attempts} attempts starting from {start_port}")


def main():
    """Main entry point for xagent command."""
    parser = argparse.ArgumentParser(description="xAgent Configuration UI")
    parser.add_argument("--host", default="0.0.0.0", help="Host to bind to")
    parser.add_argument("--port", type=int, default=8502, help="Port to bind to")
    parser.add_argument("--browser", action="store_true", help="Open browser automatically")
    
    args = parser.parse_args()
    
    try:
        # Find an available port starting from the requested port
        available_port = find_available_port(args.host, args.port)
        
        if available_port != args.port:
            print(f"Port {args.port} is already in use, using port {available_port} instead")
        
        # Get the config UI file path
        config_ui_path = Path(__file__).parent / "config_ui.py"
        
        # Build streamlit command
        cmd = [
            "streamlit", "run", str(config_ui_path),
            "--server.address", args.host,
            "--server.port", str(available_port),
            "--server.headless", "true" if not args.browser else "false"
        ]
        
        if not args.browser:
            cmd.extend(["--browser.gatherUsageStats", "false"])
        
        print(f"Starting xAgent Configuration UI on {args.host}:{available_port}")
        print(f"Access the interface at: http://{args.host}:{available_port}")
        
        # Run streamlit
        subprocess.run(cmd)
        
    except KeyboardInterrupt:
        print("\nShutting down xAgent Configuration UI...")
        sys.exit(0)
    except Exception as e:
        print(f"Failed to start configuration UI: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



import redis.asyncio as redis
from typing import List, Optional, Final

from ..schemas import Message

import os
import logging
from urllib.parse import quote
from redis.exceptions import RedisError

logger = logging.getLogger(__name__)


class MessageDB:
    """
    Stores all message history using Redis as the backend.

    All message histories are isolated with a unified prefix (chat:), support multiple sessions, and allow message trimming and expiration.
    Main features:
    - Store message history by user/session
    - Support message append, retrieval, trimming, and expiration
    - Unified Redis key encapsulation for easier maintenance
    """
    MSG_PREFIX: Final[str] = "chat"
    DEFAULT_TTL: Final[int] = 2592000  # 30 days

    def __init__(self, redis_url: str = None, *, sanitize_keys: bool = False):
        """
        Initialize MessageDB instance and connect to Redis.
        
        Args:
            redis_url (str, optional): Redis connection URL. Uses parameter first, 
                otherwise reads from REDIS_URL environment variable.
            sanitize_keys (bool, optional): Whether to URL-encode keys for safety. Defaults to False.
        
        Raises:
            ValueError: Redis connection information not provided.
        """
        url = redis_url or os.environ.get("REDIS_URL")
        if not url:
            raise ValueError("REDIS_URL not set in environment or not provided as argument")
        self.redis_url = url
        self.r: Optional[redis.Redis] = None
        self._sanitize_keys = sanitize_keys

    async def _get_client(self) -> redis.Redis:
        """
        Get or create async Redis client with sane defaults.
        
        Returns:
            redis.Redis: Configured Redis client instance.
            
        Raises:
            Exception: If Redis connection fails during initial ping.
        """
        if self.r is None:
            # Use reasonable connection parameters to improve stability and robustness
            self.r = redis.Redis.from_url(
                self.redis_url,
                decode_responses=True,
                health_check_interval=30,
                socket_connect_timeout=5,
                socket_timeout=5,
                retry_on_timeout=True,
                client_name="xagent-message-db",
            )
            try:
                await self.r.ping()
            except Exception as e:
                logger.error("Redis initial ping failed: %s", e)
                raise
        return self.r

    def _make_key(self, user_id: str, session_id: Optional[str] = None) -> str:
        """
        Generate Redis key.
        
        Args:
            user_id (str): User ID.
            session_id (str, optional): Session ID.
            
        Returns:
            str: Redis key in format 'chat:<user_id>' or 'chat:<user_id>:<session_id>'.
        """
        if self._sanitize_keys:
            user_id = quote(user_id, safe="-._~")
            if session_id:
                session_id = quote(session_id, safe="-._~")
        
        if session_id:
            return f"{self.MSG_PREFIX}:{user_id}:{session_id}"
        return f"{self.MSG_PREFIX}:{user_id}"

    async def add_messages(
        self,
        user_id: str,
        messages: Message | List[Message],
        session_id: Optional[str] = None,
        ttl: int = 2592000,
        *,
        max_len: Optional[int] = None,
        reset_ttl: bool = True,
    ) -> None:
        """
        Append one or more messages to message history and set expiration time.
        
        Args:
            user_id (str): User ID.
            messages (Message | List[Message]): Message object or list of message objects.
            session_id (str, optional): Session ID.
            ttl (int): Expiration time in seconds, defaults to 30 days.
            max_len (Optional[int]): If provided, trim history to this maximum length after appending.
            reset_ttl (bool): Whether to refresh expiration time (sliding expiration). Defaults to True.
            
        Raises:
            ValueError: If ttl or max_len are invalid.
            RedisError: If Redis operation fails.
        """
        if ttl is not None and ttl <= 0:
            raise ValueError("ttl must be a positive integer when provided")
        if max_len is not None and max_len <= 0:
            raise ValueError("max_len must be a positive integer when provided")

        client = await self._get_client()
        key = self._make_key(user_id, session_id)
        
        if not isinstance(messages, list):
            messages = [messages]
        if not messages:
            return

        # Use pipeline to batch multiple round trips and maintain atomicity
        try:
            async with client.pipeline(transaction=False) as pipe:
                pipe.rpush(key, *(m.model_dump_json() for m in messages))
                if max_len is not None:
                    pipe.ltrim(key, -max_len, -1)
                if reset_ttl and ttl is not None:
                    pipe.expire(key, ttl)
                await pipe.execute()
        except RedisError as e:
            logger.error("Failed to add messages for key %s: %s", key, e)
            raise
        

    async def get_messages(self, user_id: str, session_id: Optional[str] = None, count: int = 20) -> List[Message]:
        """
        Get message history, retrieving the most recent `count` messages in reverse order.
        
        Args:
            user_id (str): User ID.
            session_id (str, optional): Session ID.
            count (int): Number of messages to retrieve, defaults to 20.
            
        Returns:
            List[Message]: List of message objects, sorted in chronological order.
            
        Raises:
            RedisError: If Redis operation fails.
        """
        if count <= 0:
            return []
        
        client = await self._get_client()
        key = self._make_key(user_id, session_id)
        try:
            raw_msgs = await client.lrange(key, -count, -1)
        except RedisError as e:
            logger.error("Failed to get messages for key %s: %s", key, e)
            raise
        
        messages: List[Message] = []
        for i, m in enumerate(raw_msgs):
            try:
                messages.append(Message.model_validate_json(m))
            except Exception as e:
                # Control log size to avoid printing overly long strings
                preview = m[:120] + ("..." if len(m) > 120 else "")
                logger.warning(
                    "Skip invalid message at index %d for key %s: %s | payload preview=%r",
                    i, key, e, preview,
                )
        return messages

    async def trim_history(self, user_id: str, session_id: Optional[str] = None, max_len: int = 200) -> None:
        """
        Trim message history, keeping only the most recent `max_len` messages.
        
        Args:
            user_id (str): User ID.
            session_id (str, optional): Session ID.
            max_len (int): Maximum number of messages to retain, defaults to 200.
            
        Raises:
            ValueError: If max_len is not positive.
            RedisError: If Redis operation fails.
        """
        if max_len <= 0:
            raise ValueError("max_len must be a positive integer")
        
        client = await self._get_client()
        key = self._make_key(user_id, session_id)
        try:
            await client.ltrim(key, -max_len, -1)
        except RedisError as e:
            logger.error("Failed to trim history for key %s: %s", key, e)
            raise

    async def set_expire(self, user_id: str, session_id: Optional[str] = None, ttl: int = 2592000) -> None:
        """
        Set expiration time for message history.
        
        Args:
            user_id (str): User ID.
            session_id (str, optional): Session ID.
            ttl (int): Expiration time in seconds, defaults to 30 days.
            
        Raises:
            ValueError: If ttl is not positive.
            RedisError: If Redis operation fails.
        """
        if ttl <= 0:
            raise ValueError("ttl must be a positive integer")
        
        client = await self._get_client()
        key = self._make_key(user_id, session_id)
        try:
            await client.expire(key, ttl)
        except RedisError as e:
            logger.error("Failed to set expire for key %s: %s", key, e)
            raise

    async def clear_history(self, user_id: str, session_id: Optional[str] = None) -> None:
        """
        Clear message history.
        
        Args:
            user_id (str): User ID.
            session_id (str, optional): Session ID.
            
        Raises:
            RedisError: If Redis operation fails.
        """
        client = await self._get_client()
        key = self._make_key(user_id, session_id)
        try:
            await client.delete(key)
        except RedisError as e:
            logger.error("Failed to clear history for key %s: %s", key, e)
            raise

    async def pop_message(self, user_id: str, session_id: Optional[str] = None) -> Optional[Message]:
        """
        Remove and return the last non-tool_result message. If the last message is a tool_result,
        automatically continue popping until a non-tool_result message is found or the list is empty.
        
        Args:
            user_id (str): User ID.
            session_id (str, optional): Session ID.
            
        Returns:
            Optional[Message]: The removed message object (non-tool_result), or None if empty.
            
        Raises:
            RedisError: If Redis operation fails.
        """
        client = await self._get_client()
        key = self._make_key(user_id, session_id)
        
        while True:
            try:
                raw_msg = await client.rpop(key)
            except RedisError as e:
                logger.error("Failed to pop message for key %s: %s", key, e)
                raise

            if raw_msg is None:
                return None

            try:
                msg = Message.model_validate_json(raw_msg)
            except Exception as e:
                preview = raw_msg[:120] + ("..." if len(raw_msg) > 120 else "")
                logger.warning("Skip invalid popped message for key %s: %s | payload preview=%r", key, e, preview)
                # Continue to next message if this is a tool_result
                continue

            if not msg.tool_call:
                return msg
            # If it's a tool_result, continue the loop

    async def close(self) -> None:
        """Close the Redis connection."""
        if self.r:
            try:
                await self.r.aclose()
            finally:
                self.r = None

    async def __aenter__(self):
        """Async context manager entry."""
        await self._get_client()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """Async context manager exit."""
        await self.close()


import os
import time
import redis
from dotenv import load_dotenv

load_dotenv(override=True)

from .vocabulary_schema import VocabularyRecord

class VocabularyDB:
    """
    VocabularyDB
    -------------------
    Redis-backed vocabulary database for user vocabulary records.
    所有词汇记录都以统一前缀（vocab:）隔离，便于多用户、多单词管理。
    主要功能：
    - 按用户/单词存储词汇记录
    - 支持词汇保存、获取、删除、批量获取、熟悉度调整、extra 字段设置
    - Redis key 统一封装，便于维护
    """
    VOCAB_PREFIX: str = "vocab"

    def __init__(self, redis_url: str = None):
        """
        初始化 VocabularyDB 实例，连接 Redis。
        Args:
            redis_url (str, optional): Redis 连接 URL。优先使用参数，否则读取环境变量 REDIS_URL。
        Raises:
            ValueError: 未提供 Redis 连接信息。
            ConnectionError: Redis 连接失败。
        """
        url = redis_url or os.environ.get("REDIS_URL")
        if not url:
            raise ValueError("REDIS_URL not set in environment or not provided as argument")
        try:
            self.client: redis.Redis = redis.Redis.from_url(url)
            self.client.ping()
        except Exception as e:
            raise ConnectionError(f"Failed to connect to Redis at {url}: {e}")
        
    def _make_key(self, user_id: str, word: str) -> str:
        """
        生成 Redis key，格式为 'vocab:<user_id>:<word>'。
        Args:
            user_id (str): 用户 ID / User ID.
            word (str): 单词 / Word.
        Returns:
            str: Redis key 字符串 / Redis key string.
        """
        return f"{self.VOCAB_PREFIX}:{user_id.lower()}:{word.lower()}"

    def save_vocabulary(self, vocab: VocabularyRecord) -> bool:
        """
        保存或更新一个词汇记录到 Redis，并更新时间戳。
        Args:
            vocab (VocabularyRecord): 词汇记录实例 / VocabularyRecord instance.
        Returns:
            bool: 总是 True / Always True.
        """
        vocab.update_timestamp = time.time()
        key = self._make_key(vocab.user_id, vocab.word)
        value = vocab.model_dump_json()
        self.client.set(key, value)
        return True

    def get_vocabulary(self, user_id: str, word: str, reduce_familiarity: bool = False) -> VocabularyRecord | None:
        """
        获取指定用户的某个单词的词汇记录。
        Args:
            user_id (str): 用户 ID / User ID.
            word (str): 单词 / Word.
            reduce_familiarity (bool): 是否减少熟悉度（如为 True，熟悉度减 1，最小为 0，并更新时间戳）/ If True, decrease familiarity by 1 (min 0) and update timestamp.
        Returns:
            VocabularyRecord | None: 词汇记录或 None / The vocabulary record or None if not found.
        """
        key = self._make_key(user_id, word)
        value = self.client.get(key)
        if value:
            vocab = VocabularyRecord.model_validate_json(value)
            if reduce_familiarity:
                vocab.familiarity = max(0, vocab.familiarity - 1)
                vocab.update_timestamp = time.time()
                self.client.set(key, vocab.model_dump_json())  # 更新时间戳
            return vocab
        return None

    def delete_vocabulary(self, user_id: str, word: str) -> int:
        """
        删除指定用户的某个单词的词汇记录。
        Args:
            user_id (str): 用户 ID / User ID.
            word (str): 单词 / Word.
        Returns:
            int: 删除的 key 数量（0 或 1）/ Number of keys deleted (0 or 1).
        """
        key = self._make_key(user_id, word)
        return self.client.delete(key)

    def get_all_words_by_user(self, user_id: str, exclude_known: bool = False) -> list[VocabularyRecord]:
        """
        获取指定用户的所有词汇记录。
        Args:
            user_id (str): 用户 ID / User ID.
            exclude_known (bool): 是否排除熟悉度为 10 的单词 / If True, exclude words with familiarity == 10.
        Returns:
            list[VocabularyRecord]: 词汇记录列表 / List of vocabulary records.
        """
        pattern = f"{self.VOCAB_PREFIX}:{user_id}:*"
        result = []
        for key in self.client.scan_iter(pattern):
            value = self.client.get(key)
            if value:
                vocab = VocabularyRecord.model_validate_json(value)
                if exclude_known and getattr(vocab, "familiarity", None) == 10:
                    continue
                result.append(vocab)
        return result

    def update_familiarity(self, user_id: str, word: str, delta: int) -> VocabularyRecord | None:
        """
        调整指定单词的熟悉度（加/减），范围限定在 0-10。
        Args:
            user_id (str): 用户 ID / User ID.
            word (str): 单词 / Word.
            delta (int): 增量（正负均可）/ Increment (positive or negative).
        Returns:
            VocabularyRecord | None: 更新后的词汇记录或 None / Updated vocabulary record or None if not found.
        """
        vocab = self.get_vocabulary(user_id, word)
        if not vocab:
            return None
        vocab.familiarity = max(0, min(10, vocab.familiarity + delta))
        vocab.update_timestamp = time.time()
        self.save_vocabulary(vocab)
        return vocab

    def set_extra(self, user_id: str, word: str, extra: dict[str, str], mode: str = "overwrite") -> VocabularyRecord | None:
        """
        设置指定单词的 extra 字段。
        Args:
            user_id (str): 用户 ID / User ID.
            word (str): 单词 / Word.
            extra (dict[str, str]): 要设置或合并的 extra 字典 / Extra dictionary to set or merge.
            mode (str): 'overwrite'（完全覆盖）或 'add'（合并）/ 'overwrite' (replace) or 'add' (merge).
        Returns:
            VocabularyRecord | None: 更新后的词汇记录或 None / Updated vocabulary record or None if not found.
        """
        vocab = self.get_vocabulary(user_id, word)
        if not vocab:
            return None
        if mode == "add":
            if vocab.extra is None:
                vocab.extra = extra.copy()
            else:
                vocab.extra.update(extra)
        else:  # overwrite
            vocab.extra = extra.copy()
        vocab.update_timestamp = time.time()
        self.save_vocabulary(vocab)
        return vocab

    def clear_all(self) -> None:
        """
        清空数据库中的所有词汇记录。
        Returns:
            None
        """
        pattern = f"{self.VOCAB_PREFIX}:*"
        keys = list(self.client.scan_iter(pattern))
        if keys:
            self.client.delete(*keys)
        return True



import os
import yaml
import uvicorn
import argparse
from typing import Optional, Dict, Any
from fastapi import FastAPI, HTTPException
from pydantic import BaseModel
import json
from fastapi.responses import StreamingResponse
from dotenv import load_dotenv

load_dotenv(override=True)

from ..core.base import BaseAgentRunner
from ..core.session import Session


class AgentInput(BaseModel):
    """Request body for chat endpoint."""
    user_id: str
    session_id: str
    user_message: str
    image_source: Optional[str] = None
    # Enable server-side streaming when true
    stream: Optional[bool] = False


class ClearSessionInput(BaseModel):
    """Request body for clear session endpoint."""
    user_id: str
    session_id: str


class HTTPAgentServer(BaseAgentRunner):
    """HTTP Agent Server for xAgent."""
    
    def __init__(self, config_path: Optional[str] = None, toolkit_path: Optional[str] = None):
        """
        Initialize HTTPAgentServer.
        
        Args:
            config_path: Path to configuration file (if None, uses default configuration)
            toolkit_path: Path to toolkit directory (if None, no additional tools will be loaded)
        """
        # Initialize the base agent runner
        super().__init__(config_path, toolkit_path)
        
        # Initialize FastAPI app
        self.app = self._create_app()
        
    def _create_app(self) -> FastAPI:
        """Create and configure FastAPI application."""
        app = FastAPI(
            title="xAgent HTTP Agent Server",
            description="HTTP API for xAgent conversational AI",
            version="1.0.0"
        )
        
        # Add routes
        self._add_routes(app)
        
        return app
    
    def _add_routes(self, app: FastAPI) -> None:
        """Add API routes to the FastAPI application."""
        
        @app.get("/health")
        async def health_check():
            """Health check endpoint."""
            return {"status": "healthy", "service": "xAgent HTTP Server"}
        
        @app.post("/chat")
        async def chat(input_data: AgentInput):
            """
            Chat endpoint for agent interaction.
            
            Args:
                input_data: User input containing message and metadata
                
            Returns:
                Agent response or streaming SSE when input_data.stream is True
            """
            try:
                session = Session(
                    user_id=input_data.user_id,
                    session_id=input_data.session_id,
                    message_db=self.message_db
                )
                
                # Streaming mode via Server-Sent Events
                if input_data.stream:
                    async def event_generator():
                        try:
                            response = await self.agent(
                                user_message=input_data.user_message,
                                session=session,
                                image_source=input_data.image_source,
                                stream=True
                            )
                            # If the agent returns an async generator, stream deltas
                            if hasattr(response, "__aiter__"):
                                async for delta in response:
                                    # Send as SSE data frames
                                    yield f"data: {json.dumps({'delta': delta})}\n\n"
                                # Signal completion
                                yield "data: [DONE]\n\n"
                            else:
                                # Fallback when no generator is returned
                                # Handle structured output properly
                                if hasattr(response, 'model_dump'):  # Pydantic BaseModel
                                    yield f"data: {json.dumps({'message': response.model_dump()})}\n\n"
                                else:  # String response
                                    yield f"data: {json.dumps({'message': str(response)})}\n\n"
                                yield "data: [DONE]\n\n"
                        except Exception as e:
                            # Stream error as SSE, client can handle gracefully
                            yield f"data: {json.dumps({'error': str(e)})}\n\n"
                            yield "data: [DONE]\n\n"
                    return StreamingResponse(event_generator(), media_type="text/event-stream")
                
                # Non-streaming mode (default)
                response = await self.agent(
                    user_message=input_data.user_message,
                    session=session,
                    image_source=input_data.image_source
                )
                
                # Handle different response types properly
                if hasattr(response, 'model_dump'):  # Pydantic BaseModel
                    return {"reply": response.model_dump()}
                else:  # String response
                    return {"reply": str(response)}
                
            except Exception as e:
                raise HTTPException(status_code=500, detail=f"Agent processing error: {str(e)}")
        
        @app.post("/clear_session")
        async def clear_session(input_data: ClearSessionInput):
            """
            Clear session data endpoint.
            
            Args:
                input_data: Contains user_id and session_id to clear
                
            Returns:
                Success confirmation
            """
            try:
                session = Session(
                    user_id=input_data.user_id,
                    session_id=input_data.session_id,
                    message_db=self.message_db
                )
                
                await session.clear_session()
                
                return {"status": "success", "message": f"Session {input_data.session_id} for user {input_data.user_id} cleared"}
                
            except Exception as e:
                raise HTTPException(status_code=500, detail=f"Failed to clear session: {str(e)}")

    
    def run(self, host: str = None, port: int = None) -> None:
        """
        Run the HTTP server.
        
        Args:
            host: Host to bind to
            port: Port to bind to
        """
        server_cfg = self.config.get("server", {})
        
        # Use provided args or fall back to config defaults
        host = host or server_cfg.get("host", "0.0.0.0")
        port = port or server_cfg.get("port", 8010)
        
        print(f"Starting xAgent HTTP Server on {host}:{port}")
        print(f"Agent: {self.agent.name}")
        print(f"Model: {self.agent.model}")
        print(f"Tools: {len(self.agent.tools)} loaded")
        
        uvicorn.run(
            self.app,
            host=host,
            port=port,
        )


# Global server instance for uvicorn module loading
_server_instance = None


def get_app() -> FastAPI:
    """Get the FastAPI app instance for uvicorn."""
    global _server_instance
    if _server_instance is None:
        # Use default configuration when used as module
        _server_instance = HTTPAgentServer()
    return _server_instance.app


def get_app_lazy() -> FastAPI:
    """Lazy initialization for global app variable."""
    return get_app()


# For backward compatibility - use lazy initialization to avoid import-time errors
app = None  # Will be initialized when first accessed


def main():
    """Main entry point for xagent-server command."""
    parser = argparse.ArgumentParser(description="xAgent HTTP Server")
    parser.add_argument("--config", default=None, help="Config file path (if not specified, uses default configuration)")
    parser.add_argument("--toolkit_path", default=None, help="Toolkit directory path (if not specified, no additional tools will be loaded)")
    parser.add_argument("--host", default=None, help="Host to bind to")
    parser.add_argument("--port", type=int, default=None, help="Port to bind to")
    
    args = parser.parse_args()
    
    try:
        server = HTTPAgentServer(config_path=args.config, toolkit_path=args.toolkit_path)
        server.run(host=args.host, port=args.port)
    except Exception as e:
        print(f"Failed to start server: {e}")
        raise


if __name__ == "__main__":
    main()


"""Utilities for xAgent package."""

from .tool_decorator import function_tool
from .mcp_convertor import MCPTool
from .image_upload import upload_image

__all__ = ["function_tool", "MCPTool", "upload_image"]



import os
import boto3
from botocore.exceptions import NoCredentialsError
from dotenv import load_dotenv
import uuid

load_dotenv(override=True)

s3 = boto3.client(
    's3',
    aws_access_key_id=os.getenv("AWS_ACCESS_KEY_ID"),
    aws_secret_access_key=os.getenv("AWS_SECRET_ACCESS_KEY"),
    region_name=os.getenv("AWS_REGION")
)

def upload_image(file_path: str) -> str:
    bucket = os.getenv("BUCKET_NAME")
    name = f"upload_{uuid.uuid4().hex[:8]}_{os.path.basename(file_path)}"
    try:
        s3.upload_file(file_path, bucket, name)
        return f"https://{bucket}.s3.amazonaws.com/{name}"
    except (FileNotFoundError, NoCredentialsError) as e:
        print("File not found or credentials error:", e)
        return None
    

if __name__ == "__main__":
    # Example usage of the upload_image function
    file_path = "tests/assets/test_image.png"  # Replace with your image file path
    uploaded_url = upload_image(file_path)
    if uploaded_url:
        print("Image uploaded successfully:", uploaded_url)
    else:
        print("Image upload failed.")


import random
from fastmcp import FastMCP

mcp = FastMCP(name="Test MCP Server")

@mcp.tool
def roll_dice(n_dice: int) -> list[int]:
    """Roll `n_dice` 6-sided dice and return the results."""
    return [random.randint(1, 6) for _ in range(n_dice)]

@mcp.tool(enabled=False)
def add_numbers(a: int, b: int) -> int:
    """Add two numbers and return the result."""
    return a + b

@mcp.tool(
    name="get_user_details",
    exclude_args=["user_id"]
)
def get_user_details(user_id: str = None) -> str:
    """Retrieve user details based on user_id."""
    # user_id will be injected by the server, not provided by the LLM
    return "current user is 31 years old, lives in Hangzhou, China"

@mcp.resource("data://config")
def get_config() -> dict:
    """Provides the application configuration."""
    return {"theme": "dark", "version": "1.0"}

@mcp.prompt
def analyze_data(data_points: list[float]) -> str:
    """Creates a prompt asking for analysis of numerical data."""
    formatted_data = ", ".join(str(point) for point in data_points)
    return f"Please analyze these data points: {formatted_data}"

if __name__ == "__main__":
    mcp.run(transport="http", port=8000)


from ..core.agent import Agent
from ..core.session import Session
from ..schemas import Message

class SharedContext:
    """A shared context for managing data and chat history bewteen agents in a swarm."""
    def __init__(self):
        self.histories = {}

class Swarm:
    """
    A Swarm is a system for orchestrating multiple collaborative agents, enabling them to work together as a coordinated team to tackle complex tasks.
    """
    def __init__(self, agents: list[Agent]):
        self.agents = agents

    async def __call__(self, task: Message | str, *args, **kwds):
        return await self.invoke(task, *args, **kwds)

    async def invoke(self, task: Message | str):
        pass




import sys
import os
import time
from typing import Optional
import streamlit as st
import asyncio
import re
import tempfile
import uuid
import httpx
import json

# 页面配置
st.set_page_config(
    page_title="对话测试",
    page_icon="🤖",
    layout="wide",
    initial_sidebar_state="expanded"
)

class AgentHTTPClient:
    """HTTP客户端用于与xAgent服务通信"""
    
    def __init__(self, base_url: str = "http://localhost:8010"):
        self.base_url = base_url.rstrip('/')
        
    async def chat(self, user_message: str, user_id: str, session_id: str, image_source: Optional[str] = None):
        """发送聊天消息到Agent服务"""
        try:
            payload = {
                "user_id": user_id,
                "session_id": session_id,
                "user_message": user_message
            }
            if image_source:
                payload["image_source"] = image_source
            
            async with httpx.AsyncClient(timeout=httpx.Timeout(connect=10.0, read=180.0, write=10.0)) as client:
                response = await client.post(
                    f"{self.base_url}/chat",
                    json=payload
                )
            
            if response.status_code == 200:
                result = response.json()
                return result.get("reply", "")
            else:
                raise Exception(f"HTTP {response.status_code}: {response.text}")
                
        except httpx.RequestError as e:
            raise Exception(f"网络请求失败: {str(e)}")
    
    async def chat_stream(self, user_message: str, user_id: str, session_id: str, image_source: Optional[str] = None):
        """发送聊天消息到Agent服务（流式输出）"""
        try:
            payload = {
                "user_id": user_id,
                "session_id": session_id,
                "user_message": user_message,
                "stream": True
            }
            if image_source:
                payload["image_source"] = image_source
            
            async with httpx.AsyncClient(timeout=120.0) as client:
                async with client.stream(
                    "POST",
                    f"{self.base_url}/chat",
                    json=payload,
                    headers={"Accept": "text/event-stream"}
                ) as response:
                    
                    if response.status_code != 200:
                        raise Exception(f"HTTP {response.status_code}: {await response.aread()}")
                    
                    buffer = ""
                    async for chunk in response.aiter_bytes():
                        if chunk:
                            buffer += chunk.decode('utf-8')
                            lines = buffer.split('\n')
                            buffer = lines[-1]  # 保留最后一行可能不完整的数据
                            
                            for line in lines[:-1]:
                                line = line.strip()
                                if line.startswith('data: '):
                                    data_str = line[6:]  # 移除 'data: ' 前缀
                                    if data_str == '[DONE]':
                                        return
                                    try:
                                        data = json.loads(data_str)
                                        if 'delta' in data:
                                            yield data['delta']
                                        elif 'message' in data:
                                            yield data['message']
                                        elif 'error' in data:
                                            raise Exception(data['error'])
                                    except json.JSONDecodeError:
                                        continue
                                        
        except httpx.RequestError as e:
            raise Exception(f"网络请求失败: {str(e)}")
    
    async def clear_session(self, user_id: str, session_id: str):
        """清空会话历史"""
        try:
            payload = {
                "user_id": user_id,
                "session_id": session_id
            }
            
            async with httpx.AsyncClient(timeout=10.0) as client:
                response = await client.post(
                    f"{self.base_url}/clear_session",
                    json=payload
                )
            
            if response.status_code == 200:
                return True
            else:
                raise Exception(f"HTTP {response.status_code}: {response.text}")
                
        except httpx.RequestError as e:
            raise Exception(f"网络请求失败: {str(e)}")
    
    def health_check(self):
        """检查服务健康状态"""
        try:
            with httpx.Client(timeout=5.0) as client:
                response = client.get(f"{self.base_url}/health")
                return response.status_code == 200
        except:
            return False

# 初始化 Session State
def init_session_state():
    """初始化 Streamlit session state"""
    if "messages" not in st.session_state:
        st.session_state.messages = []
    
    if "http_client" not in st.session_state:
        st.session_state.http_client = None
    
    if "user_id" not in st.session_state:
        st.session_state.user_id = "streamlit_user"
    
    if "session_id" not in st.session_state:
        st.session_state.session_id = str(uuid.uuid4())
    
    if "agent_server_url" not in st.session_state:
        # 从环境变量读取初始值，如果没有则使用默认值
        st.session_state.agent_server_url = os.getenv("XAGENT_SERVER_URL", "http://localhost:8010")
    
    if "show_image_upload" not in st.session_state:
        st.session_state.show_image_upload = False
    
    if "enable_streaming" not in st.session_state:
        st.session_state.enable_streaming = True

def create_http_client(agent_server_url: str):
    """创建 HTTP 客户端实例"""
    return AgentHTTPClient(base_url=agent_server_url)

def render_markdown_with_img_limit(content: str, max_width: int = 400):
    """
    将 markdown 图片语法替换为带最大宽度限制的 HTML img 标签
    """
    def replacer(match):
        alt = match.group(1)
        url = match.group(2)
        return f'<img src="{url}" alt="{alt}" style="max-width:{max_width}px;">'
    # 匹配 ![alt](url)
    pattern = r'!\[([^\]]*)\]\(([^)]+)\)'
    return re.sub(pattern, replacer, content)

def display_chat_history():
    """显示聊天历史"""
    for message in st.session_state.messages:
        with st.chat_message(message["role"]):
            content = message["content"]
            
            # 检查是否为字典/JSON对象
            if isinstance(content, dict):
                # 如果是JSON对象，使用st.json展示
                st.json(content)
            elif isinstance(content, str):
                # 判断是否为 base64 图片 markdown
                if content.startswith("![generated image](data:image/png;base64,"):
                    prefix = "![generated image]("
                    suffix = ")"
                    img_url = content[len(prefix):-len(suffix)]
                    st.markdown(
                        f'<img src="{img_url}" style="max-width:400px;">',
                        unsafe_allow_html=True
                    )
                else:
                    # 新增：对所有 markdown 内容做图片宽度限制
                    st.markdown(render_markdown_with_img_limit(content), unsafe_allow_html=True)
            else:
                # 其他类型，转换为字符串显示
                st.text(str(content))
                
            if "timestamp" in message:
                st.caption(f"时间: {time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(message['timestamp']))}")

def main():
    """主函数"""
    init_session_state()
    
    # 侧边栏配置
    with st.sidebar:
        st.title("对话配置")
        
        # 服务器配置
        st.subheader("Agent 服务器设置")
        agent_server_url = st.text_input("Agent 服务器地址", value=st.session_state.agent_server_url)
        
        # 检查服务器连接状态
        if st.session_state.http_client:
            if st.session_state.http_client.health_check():
                st.success("✅ 服务器连接正常")
            else:
                st.error("❌ 服务器连接失败")
        
        # 用户配置
        st.subheader("用户设置")
        user_id = st.text_input("用户ID", value=st.session_state.user_id)
        session_id = st.text_input("会话ID (可选)", value=st.session_state.session_id or "")
        
        # 新增：图片上传模块显示控制
        st.subheader("界面设置")
        show_image_upload = st.checkbox("显示图片上传模块", value=st.session_state.show_image_upload)
        if show_image_upload != st.session_state.show_image_upload:
            st.session_state.show_image_upload = show_image_upload
            st.rerun()
        
        # 新增：流式输出控制
        enable_streaming = st.checkbox("启用流式输出", value=st.session_state.enable_streaming)
        if enable_streaming != st.session_state.enable_streaming:
            st.session_state.enable_streaming = enable_streaming
            st.rerun()
        
        # 应用配置按钮
        if st.button("应用配置", type="primary"):
            st.session_state.user_id = user_id
            st.session_state.session_id = session_id if session_id else None
            st.session_state.agent_server_url = agent_server_url
            
            # 重新创建 HTTP 客户端
            try:
                http_client = create_http_client(agent_server_url)
                st.session_state.http_client = http_client
                st.success("配置已应用！")
            except Exception as e:
                st.error(f"配置失败: {str(e)}")
        
        # 清空历史按钮
        if st.button("清空对话历史", type="secondary"):
            if st.session_state.http_client:
                try:
                    # 调用清空会话的 HTTP 接口
                    success = asyncio.run(st.session_state.http_client.clear_session(
                        st.session_state.user_id, 
                        st.session_state.session_id
                    ))
                    if success:
                        st.session_state.messages = []
                        st.success("对话历史已清空！")
                        st.rerun()
                except Exception as e:
                    st.error(f"清空历史失败: {str(e)}")
        
        
    # 主界面
    st.title("Conversational AI")

    # 初始化 HTTP 客户端（如果还没有）
    if st.session_state.http_client is None:
        try:
            http_client = create_http_client(st.session_state.agent_server_url)
            st.session_state.http_client = http_client
        except Exception as e:
            st.error(f"初始化 HTTP 客户端失败: {str(e)}")
            st.stop()
    
    # 显示聊天历史
    display_chat_history()
    
    # 聊天输入和图片上传移动到底部
    image_path = None
    image_bytes = None
    prompt = None
    
    with st._bottom:
        if st.session_state.show_image_upload:
            left_col, right_col = st.columns(2)
            with left_col:
                st.subheader("对话输入")
                prompt = st.chat_input("Type here your question...")
            with right_col:
                uploaded_image = st.file_uploader("上传图片（可选，支持jpg/png）", type=["jpg", "jpeg", "png"])
                if uploaded_image is not None:
                    image_bytes = uploaded_image.read()
                    # 保存到临时文件
                    suffix = "." + uploaded_image.type.split('/')[-1]
                    with tempfile.NamedTemporaryFile(delete=False, suffix=suffix) as tmp_file:
                        tmp_file.write(image_bytes)
                        image_path = tmp_file.name
                    # st.image(image_bytes, caption=None, width=50)
        else:
            prompt = st.chat_input("Type here your question...")

    # 聊天输入逻辑调整为使用 prompt
    if prompt:
        # 显示用户消息
        with st.chat_message("user"):
            # 新增：对用户输入内容做图片宽度限制
            st.markdown(render_markdown_with_img_limit(prompt), unsafe_allow_html=True)
            if image_bytes:
                st.image(image_bytes, caption="本次消息附带图片", width=200)
        
        # 添加用户消息到历史
        user_message = {
            "role": "user",
            "content": prompt,
            "timestamp": time.time()
        }
        if image_path:
            user_message["image_path"] = image_path
        st.session_state.messages.append(user_message)
        
        # 生成助手回复
        with st.chat_message("assistant"):
            try:
                if st.session_state.enable_streaming:
                    # 流式输出模式
                    reply_placeholder = st.empty()
                    full_reply = ""  # 初始化为字符串，但可能会被字典覆盖
                    first_chunk_received = False
                    
                    # 初始显示"正在思考..."
                    with reply_placeholder:
                        st.caption("正在思考...")
                    
                    async def stream_response():
                        nonlocal full_reply, first_chunk_received
                        async for chunk in st.session_state.http_client.chat_stream(
                            user_message=prompt,
                            user_id=st.session_state.user_id,
                            session_id=st.session_state.session_id,
                            image_source=image_path if image_path else None
                        ):
                            # 收到第一个chunk时，清除"正在思考..."提示
                            if not first_chunk_received:
                                first_chunk_received = True
                                reply_placeholder.empty()
                            
                            # 检查chunk类型，如果是字典则转换为字符串
                            if isinstance(chunk, dict):
                                # 如果是字典，可能是JSON响应，直接显示并结束
                                reply_placeholder.json(chunk)
                                full_reply = chunk  # 保存为字典
                                return
                            elif isinstance(chunk, str):
                                full_reply += chunk
                                # 实时更新显示
                                if full_reply.startswith("![generated image](data:image/png;base64,"):
                                    # 如果是图片，等待完整后再显示
                                    if full_reply.endswith(")"):
                                        prefix = "![generated image]("
                                        suffix = ")"
                                        img_url = full_reply[len(prefix):-len(suffix)]
                                        reply_placeholder.markdown(
                                            f'<img src="{img_url}" style="max-width:400px;">',
                                            unsafe_allow_html=True
                                        )
                                else:
                                    reply_placeholder.markdown(render_markdown_with_img_limit(full_reply), unsafe_allow_html=True)
                            else:
                                # 其他类型转换为字符串处理
                                chunk_str = str(chunk)
                                full_reply += chunk_str
                                reply_placeholder.markdown(render_markdown_with_img_limit(full_reply), unsafe_allow_html=True)
                    
                    asyncio.run(stream_response())
                    
                    reply = full_reply
                else:
                    # 非流式输出模式
                    with st.spinner("正在思考..."):
                        reply = asyncio.run(
                            st.session_state.http_client.chat(
                                user_message=prompt,
                                user_id=st.session_state.user_id,
                                session_id=st.session_state.session_id,
                                image_source=image_path if image_path else None
                            )
                        )
                        
                        # 检查是否为字典/JSON对象
                        if isinstance(reply, dict):
                            # 如果是JSON对象，使用st.json展示
                            st.json(reply)
                        elif isinstance(reply, str):
                            # 判断是否为 base64 图片 markdown
                            if reply.startswith("![generated image](data:image/png;base64,"):
                                # 提取 base64 数据
                                prefix = "![generated image]("
                                suffix = ")"
                                img_url = reply[len(prefix):-len(suffix)]
                                # 用 HTML 控制最大宽度
                                st.markdown(
                                    f'<img src="{img_url}" style="max-width:400px;">',
                                    unsafe_allow_html=True
                                )
                            else:
                                # 新增：对助手回复内容做图片宽度限制
                                st.markdown(render_markdown_with_img_limit(reply), unsafe_allow_html=True)
                        else:
                            # 其他类型，转换为字符串显示
                            st.text(str(reply))
                
                # 添加助手消息到历史
                assistant_message = {
                    "role": "assistant",
                    "content": reply,
                    "timestamp": time.time()
                }
                st.session_state.messages.append(assistant_message)
                
            except Exception as e:
                error_msg = f"生成回复时出错: {str(e)}"
                st.error(error_msg)
                
                # 添加错误消息到历史
                error_message = {
                    "role": "assistant",
                    "content": error_msg,
                    "timestamp": time.time()
                }
                st.session_state.messages.append(error_message)

if __name__ == "__main__":
    main()



import base64
import tempfile
import os
from dotenv import load_dotenv
from typing import Optional, List

load_dotenv(override=True)

from langfuse import observe
from langfuse.openai import AsyncOpenAI

from xagent.utils.tool_decorator import function_tool
from xagent.utils.image_upload import upload_image as s3_upload_image

DEFAULT_MODEL = "gpt-4o-mini"

@function_tool()
@observe()
async def web_search(search_query: str) -> str:
    "when the user wants to search the web using a search engine, use this tool"

    query = (search_query or "").strip()
    if not query:
        return ""
    
    client = AsyncOpenAI()
    response = await client.responses.create(
        model=DEFAULT_MODEL,
        tools=[{"type": "web_search_preview"},],
        input=query,
        tool_choice="required"
    )

    return (getattr(response, "output_text", "") or "").strip()


@function_tool()
@observe()
async def draw_image(prompt: str, reference_image_source: Optional[List[str]|str] = None) -> str:
    """
    when the user wants to generate an image based on a prompt, use this tool,if user mentioned some images, use them as reference
    """

    clean_prompt = (prompt or "").strip()
    if not clean_prompt:
        return ""
    
    if reference_image_source:
        if isinstance(reference_image_source, str):
            reference_image_source = [reference_image_source]
            
        content = [{"type": "input_text", "text": clean_prompt}]
        for img_url in reference_image_source:
            content.append({
                "type": "input_image",
                "image_url": img_url,
            })
        input = [{
            "role": "user",
            "content": content
        }]
    else:
        input = [{
            "role": "user",
            "content": clean_prompt
        }]
    
    client = AsyncOpenAI()
    response = await client.responses.create(
        model=DEFAULT_MODEL,
        tools=[{"type": "image_generation", "quality": "low"}],
        input=input,
        tool_choice="required"
    )

    tool_calls = getattr(response, "output", [])
    image_call = next((tc for tc in tool_calls if getattr(tc, "type", "") == "image_generation_call"), None)
    if not image_call or not getattr(image_call, "result", None):
        return ""

    image_base64 = image_call.result
    tmp_path = None
    try:
        with tempfile.NamedTemporaryFile(prefix="generated_", suffix=".png", delete=False) as tmp_file:
            tmp_file.write(base64.b64decode(image_base64))
            tmp_path = tmp_file.name
        # Upload and get URL
        url = upload_image(tmp_path)
        return url
    except Exception:
        # Fallback to inline base64 image markdown
        return f'![generated image](data:image/png;base64,{image_base64})'
    finally:
        if tmp_path and os.path.exists(tmp_path):
            try:
                os.remove(tmp_path)
            except OSError:
                pass

def upload_image(image_path: str) -> str:
    url = s3_upload_image(image_path)
    if not url:
        raise Exception("Image upload failed")
    return url


if __name__ == "__main__":

    import asyncio
    search_result = asyncio.run(web_search("What is the weather like today in hangzhou?"))
    print("Search Result:", search_result)

    image_url = asyncio.run(draw_image("A beautiful sunset over the mountains"))
    print("Generated Image URL:", image_url)


"""
Test xAgent CLI functionality.
"""

import pytest
import asyncio
from unittest.mock import patch, AsyncMock
from xagent.core.cli import CLIAgent


class TestCLIAgent:
    """Test cases for CLIAgent."""
    
    @pytest.fixture
    def mock_config(self):
        """Mock configuration data."""
        return {
            "agent": {
                "name": "TestAgent",
                "system_prompt": "You are a test agent.",
                "model": "gpt-4.1-mini",
                "tools": [],
                "mcp_servers": [],
                "use_local_session": True
            }
        }
    
    @pytest.fixture
    def cli_agent(self, mock_config):
        """Create CLI agent with mocked config."""
        with patch.object(CLIAgent, '_load_config', return_value=mock_config):
            with patch.object(CLIAgent, '_initialize_agent') as mock_agent_init:
                with patch.object(CLIAgent, '_initialize_message_db', return_value=None):
                    # Create a mock agent that can be awaited
                    mock_agent = AsyncMock()
                    mock_agent.name = "TestAgent"
                    mock_agent.model = "gpt-4.1-mini"
                    mock_agent.tools = {}
                    mock_agent.mcp_tools = {}
                    mock_agent_init.return_value = mock_agent
                    return CLIAgent(config_path="test_config.yaml")
    
    def test_cli_agent_initialization(self, cli_agent):
        """Test CLI agent initialization."""
        assert cli_agent.toolkit_path == "toolkit"
        assert cli_agent.agent.name == "TestAgent"
        assert cli_agent.agent.model == "gpt-4.1-mini"
    
    @pytest.mark.asyncio
    async def test_chat_single(self, cli_agent):
        """Test single message chat."""
        # Mock the agent to return a test response
        cli_agent.agent.return_value = "Test response"
        
        response = await cli_agent.chat_single("Test message")
        
        assert response == "Test response"
        cli_agent.agent.assert_called_once()
    
    def test_show_help(self, cli_agent, capsys):
        """Test help display."""
        cli_agent._show_help()
        captured = capsys.readouterr()
        assert "Available commands:" in captured.out
        assert "Available tools:" in captured.out


def test_main_function():
    """Test main function argument parsing."""
    # Test that main function exists and can be imported
    from xagent.core.cli import main
    assert callable(main)


if __name__ == "__main__":
    pytest.main([__file__])



import os
import yaml
import importlib.util
import sys
from typing import Optional, Dict, Any, Type, List
from dotenv import load_dotenv
from pydantic import BaseModel, create_model

from ..core.agent import Agent
from ..db.message import MessageDB
from ..tools import TOOL_REGISTRY


class BaseAgentRunner:
    """Base class for agent runners with common configuration and initialization logic."""
    
    def __init__(self, config_path: Optional[str] = None, toolkit_path: Optional[str] = None):
        """
        Initialize BaseAgentRunner.
        
        Args:
            config_path: Path to configuration file (if None, uses default configuration)
            toolkit_path: Path to toolkit directory (if None, no additional tools will be loaded)
        """
        # Load environment variables
        load_dotenv(override=True)
        
        # Persist toolkit path for dynamic loading
        self.toolkit_path = toolkit_path
        
        # Load configuration
        self.config = self._load_config(config_path)
        
        # Initialize components
        self.agent = self._initialize_agent()
        self.message_db = self._initialize_message_db()
        
    def _load_config(self, cfg_path: Optional[str]) -> Dict[str, Any]:
        """
        Load YAML configuration file.
        
        Args:
            cfg_path: Path to config file (if None, uses default configuration)
            
        Returns:
            Configuration dictionary
        """
        # If no config path provided, use default configuration
        if cfg_path is None:
            return self._get_default_config()
        
        # Check if the specified config file exists
        if os.path.isfile(cfg_path):
            with open(cfg_path, "r", encoding="utf-8") as f:
                return yaml.safe_load(f)
        else:
            # Use default configuration if file doesn't exist
            return self._get_default_config()
    
    def _get_default_config(self) -> Dict[str, Any]:
        """
        Return default configuration when no config file is found.
        
        Returns:
            Default configuration dictionary
        """
        return {
            "agent": {
                "name": "Agent",
                "system_prompt": "You are a helpful assistant. Your task is to assist users with their queries and tasks.",
                "model": "gpt-4o-mini",
                "capabilities": {
                    "tools": ["web_search"],  # Default tools
                    "mcp_servers": []  # Default MCP servers
                },
                "use_local_session": True
            },
            "server": {
                "host": "0.0.0.0",
                "port": 8010
            }
        }
    
    def _load_toolkit_registry(self, toolkit_path: Optional[str]) -> Dict[str, Any]:
        """Dynamically load TOOLKIT_REGISTRY from a toolkit directory.
        Only a directory path is supported; do not pass __init__.py.
        Returns empty dict if unavailable or on error.
        """
        if not toolkit_path:
            return {}
        try:
            # Resolve relative paths against this file's directory
            def resolve_path(p: str) -> str:
                if os.path.isabs(p):
                    return p
                if os.path.exists(p):
                    return p
                base = os.path.dirname(os.path.abspath(__file__))
                candidate = os.path.join(base, p)
                return candidate

            tp = resolve_path(toolkit_path)

            # Require a directory
            if not os.path.isdir(tp):
                return {}

            init_path = os.path.join(tp, "__init__.py")
            if not os.path.isfile(init_path):
                return {}

            # Mark as a package so relative imports inside __init__.py work
            spec = importlib.util.spec_from_file_location(
                "xagent_dynamic_toolkit",
                init_path,
                submodule_search_locations=[tp],
            )
            if spec and spec.loader:
                module = importlib.util.module_from_spec(spec)
                sys.modules["xagent_dynamic_toolkit"] = module
                spec.loader.exec_module(module)  # type: ignore[attr-defined]
                registry = getattr(module, "TOOLKIT_REGISTRY", {})
                if isinstance(registry, dict):
                    return registry
        except Exception as e:
            print(f"Warning: failed to load TOOLKIT_REGISTRY from {toolkit_path}: {e}")
        return {}
    
    def _create_output_model_from_schema(self, output_schema: Optional[Dict[str, Any]]) -> Optional[Type[BaseModel]]:
        """
        Create a dynamic Pydantic BaseModel from YAML output_schema configuration.
        
        Args:
            output_schema: Dictionary containing class_name and fields configuration
            
        Returns:
            Dynamic Pydantic BaseModel class or None if no schema provided
        """
        if not output_schema:
            return None
        
        try:
            class_name = output_schema.get("class_name", "DynamicModel")
            fields_config = output_schema.get("fields", {})
            
            if not fields_config:
                return None
            
            # Build field definitions for create_model
            field_definitions = {}
            for field_name, field_config in fields_config.items():
                field_type = field_config.get("type", "str")
                field_description = field_config.get("description", "")
                
                # Map string type names to actual Python types
                type_mapping = {
                    "str": str,
                    "int": int,
                    "float": float,
                    "bool": bool,
                    "list": list,
                    "dict": dict,
                }
                
                python_type = type_mapping.get(field_type, str)
                
                # Handle list types with items specification
                if field_type == "list" and "items" in field_config:
                    items_type = field_config["items"]
                    items_python_type = type_mapping.get(items_type, str)
                    # Create List[items_type] annotation
                    python_type = List[items_python_type]
                
                # Create field definition (type, default_value, Field(...))
                from pydantic import Field
                field_definitions[field_name] = (python_type, Field(description=field_description))
            
            # Create the dynamic model
            dynamic_model = create_model(class_name, **field_definitions)
            return dynamic_model
            
        except Exception as e:
            print(f"Warning: failed to create output model from schema: {e}")
            return None
    
    def _initialize_agent(self) -> Agent:
        """Initialize the agent with tools and configuration."""
        agent_cfg = self.config.get("agent", {})
        
        # Load tools from built-in registry and optional toolkit registry
        capabilities = agent_cfg.get("capabilities", {})
        tool_names = capabilities.get("tools", [])
        mcp_servers = capabilities.get("mcp_servers", [])
        
        # Support legacy format for backward compatibility
        if "tools" in agent_cfg and "tools" not in capabilities:
            tool_names = agent_cfg.get("tools", [])
        if "mcp_servers" in agent_cfg and "mcp_servers" not in capabilities:
            mcp_servers = agent_cfg.get("mcp_servers", [])
        
        toolkit_registry = self._load_toolkit_registry(self.toolkit_path)
        combined_registry: Dict[str, Any] = {**TOOL_REGISTRY, **toolkit_registry}
        tools = [combined_registry[name] for name in tool_names if name in combined_registry]
        
        # Process sub_agents configuration
        sub_agents = None
        if "sub_agents" in agent_cfg:
            sub_agents = []
            for agent_config in agent_cfg["sub_agents"]:
                if isinstance(agent_config, dict):
                    # Convert dict to tuple format
                    sub_agents.append((
                        agent_config.get("name", ""),
                        agent_config.get("description", ""),
                        agent_config.get("server_url", "")
                    ))
                elif isinstance(agent_config, (list, tuple)) and len(agent_config) == 3:
                    # Already in tuple format
                    sub_agents.append(tuple(agent_config))
        
        # Create output model from schema if provided
        output_type = None
        if "output_schema" in agent_cfg:
            output_type = self._create_output_model_from_schema(agent_cfg["output_schema"])

        return Agent(
            name=agent_cfg.get("name"),
            system_prompt=agent_cfg.get("system_prompt"),
            model=agent_cfg.get("model"),
            tools=tools,
            mcp_servers=mcp_servers,
            sub_agents=sub_agents,
            output_type=output_type,
        )
    
    def _initialize_message_db(self) -> Optional[MessageDB]:
        """Initialize message database based on configuration."""
        agent_cfg = self.config.get("agent", {})
        use_local_session = agent_cfg.get("use_local_session", True)
        return None if use_local_session else MessageDB()



import os
import pytest
import time
from dotenv import load_dotenv

load_dotenv(override=True)

from tools.vocabulary.vocabulary_schema import VocabularyRecord, DifficultyLevel


@pytest.fixture(scope="function")
def vocab_db():
    """Fixture: 每个测试用例前后清理 Redis，保证测试隔离。"""
    db = VocabularyDB(os.environ.get("TEST_REDIS_URL"))
    # 清理所有以 vocab: 开头的 key
    for key in db.client.keys("vocab:*"):
        db.client.delete(key)
    yield db
    for key in db.client.keys("vocab:*"):
        db.client.delete(key)

def test_save_and_get_vocabulary(vocab_db):
    """测试保存和获取单词功能。"""
    user_id = "user123"
    ts = time.time()
    vocab = VocabularyRecord(
        word="apple", 
        explanation="A fruit", 
        user_id=user_id, 
        create_timestamp=ts, 
        familiarity=3, 
        difficulty_level=DifficultyLevel.BEGINNER,
        example_sentences=["I eat an apple.", "The apple is red."],
        extra={"part_of_speech": "noun"}
    )
    assert vocab_db.save_vocabulary(vocab) is True
    result = vocab_db.get_vocabulary(user_id, "apple")
    assert isinstance(result, VocabularyRecord)
    assert result.word == "apple"
    assert result.explanation == "A fruit"
    assert result.user_id == user_id
    assert abs(result.create_timestamp - ts) < 1  # 时间戳误差容忍1秒
    assert result.extra["part_of_speech"] == "noun"
    assert result.familiarity == 3
    assert result.difficulty_level == DifficultyLevel.BEGINNER
    assert result.example_sentences == ["I eat an apple.", "The apple is red."]

def test_update_familiarity(vocab_db):
    """测试熟悉度增减和边界。"""
    user_id = "user_fam"
    vocab = VocabularyRecord(word="testword", explanation="test", user_id=user_id, create_timestamp=0, familiarity=5, extra=None)
    vocab_db.save_vocabulary(vocab)
    # 增加熟悉度
    updated = vocab_db.update_familiarity(user_id, "testword", 3)
    assert updated.familiarity == 8
    # 超过上限
    updated = vocab_db.update_familiarity(user_id, "testword", 5)
    assert updated.familiarity == 10
    # 降低熟悉度
    updated = vocab_db.update_familiarity(user_id, "testword", -7)
    assert updated.familiarity == 3
    # 低于下限
    updated = vocab_db.update_familiarity(user_id, "testword", -10)
    assert updated.familiarity == 0
    # 不存在的词
    assert vocab_db.update_familiarity(user_id, "notfound", 1) is None

def test_delete_vocabulary(vocab_db):
    """测试删除单词功能。"""
    user_id = "user_del"
    vocab = VocabularyRecord(word="banana", explanation="A fruit", user_id=user_id, create_timestamp=time.time(), familiarity=2, extra=None)
    vocab_db.save_vocabulary(vocab)
    assert vocab_db.get_vocabulary(user_id, "banana") is not None
    assert vocab_db.delete_vocabulary(user_id, "banana") == 1
    assert vocab_db.get_vocabulary(user_id, "banana") is None

def test_get_all_words_by_user(vocab_db):
    """测试获取用户所有单词功能。"""
    user_id = "user_list"
    vocab1 = VocabularyRecord(word="cat", explanation="An animal", user_id=user_id, create_timestamp=time.time(), familiarity=1, extra=None)
    vocab2 = VocabularyRecord(word="dog", explanation="Another animal", user_id=user_id, create_timestamp=time.time(), familiarity=2, extra=None)
    vocab_db.save_vocabulary(vocab1)
    vocab_db.save_vocabulary(vocab2)
    words = vocab_db.get_all_words_by_user(user_id)
    word_set = set([v.word for v in words])
    assert "cat" in word_set and "dog" in word_set
    assert all(isinstance(v, VocabularyRecord) for v in words)

def test_get_nonexistent_vocabulary(vocab_db):
    """测试获取不存在的单词返回 None。"""
    result = vocab_db.get_vocabulary("user123", "nonexistent")
    assert result is None

def test_set_extra_overwrite_and_add(vocab_db):
    """测试 set_extra 的 add/overwrite 模式。"""
    user_id = "user_extra"
    vocab = VocabularyRecord(word="extraword", explanation="extra", user_id=user_id, create_timestamp=0, familiarity=1, extra={"a": "1"})
    vocab_db.save_vocabulary(vocab)
    # 覆盖模式
    updated = vocab_db.set_extra(user_id, "extraword", {"b": "2"}, mode="overwrite")
    assert updated.extra == {"b": "2"}
    # 合并模式
    updated = vocab_db.set_extra(user_id, "extraword", {"c": "3"}, mode="add")
    assert updated.extra == {"b": "2", "c": "3"}
    # extra 为 None 时 add
    vocab2 = VocabularyRecord(word="noneword", explanation="none", user_id=user_id, create_timestamp=0, familiarity=1, extra=None)
    vocab_db.save_vocabulary(vocab2)
    updated2 = vocab_db.set_extra(user_id, "noneword", {"x": "y"}, mode="add")
    assert updated2.extra == {"x": "y"}
    # 不存在的词
    assert vocab_db.set_extra(user_id, "notfound", {"a": "b"}) is None

def test_delete_vocabulary_not_exist(vocab_db):
    """测试删除不存在的单词返回 0。"""
    user_id = "user_del2"
    assert vocab_db.delete_vocabulary(user_id, "notfound") == 0

def test_get_all_words_by_user_empty(vocab_db):
    """测试获取用户所有单词为空时返回空列表。"""
    user_id = "user_empty"
    words = vocab_db.get_all_words_by_user(user_id)
    assert words == []
    assert isinstance(words, list)


def test_vocabularydb_init_without_url(monkeypatch):
    """测试未设置 REDIS_URL 时抛出异常。"""
    monkeypatch.delenv("REDIS_URL", raising=False)
    with pytest.raises(ValueError):
        VocabularyDB(redis_url=None)

def test_vocabulary_with_default_values(vocab_db):
    """测试词汇记录的默认值。"""
    user_id = "user_defaults"
    vocab = VocabularyRecord(
        word="default", 
        explanation="Test default values", 
        user_id=user_id, 
        create_timestamp=time.time()
    )
    vocab_db.save_vocabulary(vocab)
    
    result = vocab_db.get_vocabulary(user_id, "default")
    assert result is not None
    assert result.familiarity == 0  # 默认值
    assert result.difficulty_level == DifficultyLevel.INTERMEDIATE  # 默认值
    assert result.example_sentences == []  # 默认值
    assert result.extra is None  # 默认值



import random
from fastmcp import FastMCP

from vocabulary.vocabulary import Vocabulary

mcp = FastMCP(name="MCP Server")

vocabulary = Vocabulary()

## Tools

@mcp.tool(enabled=True)
def roll_dice(n_dice: int) -> list[int]:
    """Roll `n_dice` 6-sided dice and return the results."""
    return [random.randint(1, 6) for _ in range(n_dice)]

@mcp.tool(enabled=False)
def add_numbers(a: int, b: int) -> int:
    """Add two numbers and return the result."""
    return a + b

@mcp.tool(
    name="get_user_details",
    exclude_args=["user_id"]
)
def get_user_details(user_id: str = None) -> str:
    """Retrieve user details based on user_id."""
    # user_id will be injected by the server, not provided by the LLM
    return "current user is Jun, 31 years old, lives in Hangzhou, China , hobby is reading books"

@mcp.tool(enabled=True)
def look_up_word(word: str, user_id: str) -> str:
    """when user asks for a word definition or meaning, use this tool to look it up."""
    try:
        result = vocabulary.lookup_word(word=word, user_id=user_id)
        result = f"Definition of '{word}':\n {result}"
    except Exception as e:
        result = f"Error looking up word: {str(e)}"
    return result

@mcp.tool(enabled=True)
def get_vocabulary(user_id: str, n: int) -> list[str]:
    """when user asks for vocabulary list for review or study, use this tool to retrieve it."""
    try:
        vocabularies = vocabulary.get_vocabulary(user_id=user_id, n=n, exclude_known=True)
        vocabularies = [f"Vocabulary:\n {vocab}" for vocab in vocabularies]
    except Exception as e:
        vocabularies = [f"Error retrieving vocabulary: {str(e)}"]
    return vocabularies

## Resources

@mcp.resource("data://config")
def get_config() -> dict:
    """Provides the application configuration."""
    return {"theme": "dark", "version": "1.0"}


# Prompts

@mcp.prompt
def analyze_data(data_points: list[float]) -> str:
    """Creates a prompt asking for analysis of numerical data."""
    formatted_data = ", ".join(str(point) for point in data_points)
    return f"Please analyze these data points: {formatted_data}"

def main():
    """Main entry point for xagent-mcp command."""
    import argparse
    
    parser = argparse.ArgumentParser(description="xAgent MCP Server")
    parser.add_argument("--port", type=int, default=8001, help="Port to bind to")
    parser.add_argument("--host", default="localhost", help="Host to bind to")
    parser.add_argument("--transport", default="http", help="Transport type")
    
    args = parser.parse_args()
    
    print(f"Starting xAgent MCP Server on {args.host}:{args.port}")
    mcp.run(transport=args.transport, port=args.port, host=args.host)

if __name__ == "__main__":
    main()


import inspect
import functools
from typing import get_type_hints, get_origin, get_args, Union
import sys

def python_type_to_openai_type(py_type):
    # 处理 Optional 类型 (Optional[T] 等价于 Union[T, None])
    if get_origin(py_type) is Union:
        args = get_args(py_type)
        # 如果是 Optional 类型，去掉 None 后处理
        if len(args) == 2 and type(None) in args:
            non_none_type = args[0] if args[1] is type(None) else args[1]
            return python_type_to_openai_type(non_none_type)
        # 处理联合类型，选择第一个类型作为主要类型
        return python_type_to_openai_type(args[0])
    
    # 处理泛型 List
    if get_origin(py_type) is list:
        args = get_args(py_type)
        if args:
            item_type = python_type_to_openai_type(args[0])
            return {"type": "array", "items": item_type}
        return {"type": "array", "items": {"type": "string"}}
    
    # 基础类型处理
    if py_type is int:
        return {"type": "integer"}
    if py_type is float:
        return {"type": "number"}
    if py_type is bool:
        return {"type": "boolean"}
    if py_type is str:
        return {"type": "string"}
    if py_type is list:
        return {"type": "array", "items": {"type": "string"}}
    if py_type is dict:
        return {"type": "object"}
    
    # Python 3.10+ 的联合类型语法 (X | Y)
    if sys.version_info >= (3, 10) and hasattr(py_type, '__class__') and py_type.__class__.__name__ == 'UnionType':
        args = py_type.__args__
        return python_type_to_openai_type(args[0])
    
    return {"type": "string"}

def function_tool(name: str = None, description: str = None):
    """
    将函数包装成 openai tool call 所需的规范格式，仅支持基础类型

    Args:
        name (str): 工具名称，默认为函数名
        description (str): 工具描述，默认为函数文档字符串
    Returns:
        function: 包装后的函数，具有 tool_spec 属性
    Raises:
        TypeError: 如果函数参数类型不支持
        ValueError: 如果函数没有参数或返回值类型不支持

    Example usage:
        @function_tool(name="my_tool", description="This is my tool")
        def my_tool(param1: int, param2: str):
            \"\"\"This is my tool function.\"\"\"
            return f"Received {param1} and {param2}"

        print(my_tool.tool_spec)
    """
    import asyncio
    def decorator(func):
        sig = inspect.signature(func)
        type_hints = get_type_hints(func)
        properties = {}
        required = []
        for param in sig.parameters.values():
            param_type = type_hints.get(param.name, str)
            properties[param.name] = python_type_to_openai_type(param_type)
            if param.default is param.empty:
                required.append(param.name)
        parameters = {
            "type": "object",
            "properties": properties,
            "required": required,
            "additionalProperties": False
        }
        tool = {
            "type": "function",
            "name": name or func.__name__,
            "description": description or (func.__doc__.strip() if func.__doc__ else ""),
            "parameters": parameters
        }
        # 自动包装为异步
        if asyncio.iscoroutinefunction(func):
            async_func = func
        else:
            async def async_func(*args, **kwargs):
                loop = asyncio.get_running_loop()
                partial_func = functools.partial(func, *args, **kwargs)
                return await loop.run_in_executor(None, partial_func)
            async_func.__name__ = func.__name__
            async_func.__doc__ = func.__doc__
        async_func.tool_spec = tool
        return async_func
    return decorator


import streamlit as st
import yaml
import os
import json
import subprocess
import time
import requests
from typing import Dict, Any, List, Optional
from pathlib import Path
import signal
import psutil
from datetime import datetime

class AgentConfigUI:
    """Streamlit UI for configuring and managing xAgent HTTP servers."""
    
    def __init__(self):
        self.config_dir = Path("config")
        self.config_dir.mkdir(exist_ok=True)
        self.toolkit_dir = Path("toolkit")
        self.running_servers = self._load_server_registry()
        self.running_webchats = self._load_webchat_registry()
        
    def _load_server_registry(self) -> Dict[str, Dict]:
        """Load running server registry from file."""
        registry_file = self.config_dir / "server_registry.json"
        if registry_file.exists():
            try:
                with open(registry_file, 'r') as f:
                    return json.load(f)
            except:
                return {}
        return {}
    
    def _save_server_registry(self):
        """Save running server registry to file."""
        registry_file = self.config_dir / "server_registry.json"
        with open(registry_file, 'w') as f:
            json.dump(self.running_servers, f, indent=2)
    
    def _load_webchat_registry(self) -> Dict[str, Dict]:
        """Load running web chat registry from file."""
        registry_file = self.config_dir / "webchat_registry.json"
        if registry_file.exists():
            try:
                with open(registry_file, 'r') as f:
                    return json.load(f)
            except:
                return {}
        return {}
    
    def _save_webchat_registry(self):
        """Save running web chat registry to file."""
        registry_file = self.config_dir / "webchat_registry.json"
        with open(registry_file, 'w') as f:
            json.dump(self.running_webchats, f, indent=2)
    
    def _check_server_health(self, url: str) -> bool:
        """Check if server is healthy."""
        try:
            # Handle different host formats
            if url.startswith("http://0.0.0.0:"):
                # Replace 0.0.0.0 with localhost for health check
                url = url.replace("0.0.0.0", "localhost")
            
            response = requests.get(f"{url}/health", timeout=5)
            return response.status_code == 200
        except requests.exceptions.ConnectionError:
            # Server is not ready yet
            return False
        except requests.exceptions.Timeout:
            # Server is taking too long to respond
            return False
        except Exception:
            return False
    
    def _check_prerequisites(self) -> bool:
        """Check if all prerequisites for starting a server are met."""
        # Check if OpenAI API key is set
        if not os.getenv('OPENAI_API_KEY'):
            st.error("❌ OPENAI_API_KEY environment variable is not set")
            st.info("Please set your OpenAI API key:")
            st.code("export OPENAI_API_KEY=your_api_key_here")
            return False
        
        # Check if xagent-server command is available
        try:
            result = subprocess.run(['xagent-server', '--help'], 
                                  capture_output=True, timeout=5)
            if result.returncode != 0:
                st.error("❌ xagent-server command failed")
                return False
        except FileNotFoundError:
            st.error("❌ xagent-server command not found")
            st.info("Please install xAgent: `pip install -e .`")
            return False
        except subprocess.TimeoutExpired:
            st.warning("⚠️ xagent-server command is slow to respond")
        except Exception as e:
            st.error(f"❌ Error checking xagent-server: {e}")
            return False
        
        return True
    
    def _cleanup_dead_servers(self):
        """Remove dead servers from registry."""
        dead_servers = []
        for server_id, server_info in self.running_servers.items():
            if not self._check_server_health(server_info['url']):
                # Check if process is still running
                try:
                    pid = server_info.get('pid')
                    if pid and psutil.pid_exists(pid):
                        continue
                except:
                    pass
                dead_servers.append(server_id)
        
        for server_id in dead_servers:
            del self.running_servers[server_id]
        
        if dead_servers:
            self._save_server_registry()
    
    def _check_webchat_health(self, url: str) -> bool:
        """Check if web chat is healthy."""
        try:
            response = requests.get(url, timeout=5)
            return response.status_code == 200
        except:
            return False
    
    def _cleanup_dead_webchats(self):
        """Remove dead web chats from registry."""
        dead_webchats = []
        for webchat_id, webchat_info in self.running_webchats.items():
            if not self._check_webchat_health(webchat_info['url']):
                # Check if process is still running
                try:
                    pid = webchat_info.get('pid')
                    if pid and psutil.pid_exists(pid):
                        continue
                except:
                    pass
                dead_webchats.append(webchat_id)
        
        for webchat_id in dead_webchats:
            del self.running_webchats[webchat_id]
        
        if dead_webchats:
            self._save_webchat_registry()
    
    def _find_available_port(self, start_port: int = 8501) -> int:
        """Find an available port starting from start_port."""
        import socket
        port = start_port
        while port < start_port + 100:  # Try up to 100 ports
            try:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.bind(('localhost', port))
                    return port
            except OSError:
                port += 1
        raise Exception(f"No available port found starting from {start_port}")
    
    def render_main_page(self):
        """Render the main configuration page."""
        st.set_page_config(
            page_title="xAgent Config Manager",
            page_icon="🤖",
            layout="wide"
        )
        
        st.title("🤖 xAgent Configuration Manager")
        st.markdown("Create, configure and manage xAgent HTTP servers through a visual interface.")
        
        # Sidebar for navigation
        with st.sidebar:
            st.header("Navigation")
            page = st.radio(
                "Choose a page:",
                ["Agent Configuration", "Server Management", "Running Servers"]
            )
        
        if page == "Agent Configuration":
            self.render_config_page()
        elif page == "Server Management":
            self.render_server_management()
        elif page == "Running Servers":
            self.render_running_servers()
    
    def render_config_page(self):
        """Render the agent configuration page."""
        
        col1, col2 = st.columns([2, 1])
        
        with col1:
            # Basic Agent Configuration
            st.subheader("Basic Settings")
            
            agent_name = st.text_input(
                "Agent Name",
                value="MyAgent",
                help="Unique identifier for your agent"
            )
            
            system_prompt = st.text_area(
                "System Prompt",
                value="You are a helpful AI assistant.",
                height=100,
                help="Instructions that define the agent's behavior and personality"
            )
            
            model = st.selectbox(
                "Model",
                ["gpt-4o", "gpt-4o-mini", "gpt-4.1","gpt-4.1-mini",'gpt-4.1-nano'],
                index=1,
                help="OpenAI model to use for the agent"
            )
            
            # Server Configuration
            st.subheader("Server Settings")
            
            col_host, col_port = st.columns(2)
            with col_host:
                host = st.text_input("Host", value="0.0.0.0")
            with col_port:
                port = st.number_input("Port", min_value=1000, max_value=65535, value=8010)
            
            # Tool Configuration
            st.subheader("Tools & Capabilities")
            
            # Built-in tools
            with st.expander("Built-in Tools"):
                st.markdown("Enable or disable built-in tools for your agent.")
                enable_web_search = st.checkbox("Web Search", value=True)
                enable_draw_image = st.checkbox("Image Generation", value=False)
            
            # Custom tools
            with st.expander("Custom Tools"):
                st.markdown("Configure custom tools from your toolkit directory.")
                
                toolkit_path = st.text_input(
                    "Toolkit Path",
                    value="toolkit",
                    help="Path to custom toolkit directory containing your custom tools"
                )

                custom_tools = st.text_area(
                    "Custom Tool Names (one per line)",
                    help="Enter the names of custom tools from your toolkit",
                    placeholder="calculate_square\nfetch_weather"
                )            

                
                # Add helper info for custom tools
                if custom_tools.strip():
                    st.info("💡 Make sure your custom tools are defined in the toolkit directory and registered in `__init__.py`")
                    with st.expander("Custom Tools Setup Guide"):
                        st.markdown("""
                        **Creating Custom Tools:**
                        1. Create your toolkit directory (e.g., `toolkit/`)
                        2. Add your tool functions in `.py` files
                        3. Register tools in `__init__.py`:
                        ```python
                        # toolkit/__init__.py
                        from .your_tools import calculate_square, fetch_weather
                        
                        TOOLKIT_REGISTRY = {
                            "calculate_square": calculate_square,
                            "fetch_weather": fetch_weather
                        }
                        ```
                        4. Use the `@function_tool()` decorator on your functions
                        """)
            
            # MCP Servers
            with st.expander("MCP Servers"):
                st.markdown("Configure Model Context Protocol servers for dynamic tool loading.")
                mcp_servers = st.text_area(
                    "MCP Server URLs (one per line)",
                    help="Enter URLs of MCP servers for dynamic tool loading",
                    placeholder="http://localhost:8001/mcp/\nhttp://localhost:8002/mcp/"
                )
            
            
            st.subheader("Advanced Settings")

            # Sub-agents Configuration
            with st.expander("Sub-agents (Multi-Agent System)"):
                st.markdown("Configure specialized sub-agents for hierarchical agent systems.")
                
                with st.container():
                    col_info, col_example = st.columns(2)
                    with col_info:
                        st.info("💡 **How it works:**\n- Main agent coordinates tasks\n- Sub-agents handle specialized work\n- Automatic delegation based on task type")
                    with col_example:
                        st.success("📝 **Example Use Cases:**\n- Research + Writing agents\n- Analysis + Visualization\n- Data Processing + Reporting")
                
                sub_agents_enabled = st.checkbox("Enable Sub-agents")
                
                sub_agents_config = []
                if sub_agents_enabled:
                    num_sub_agents = st.number_input("Number of Sub-agents", min_value=1, max_value=10, value=2)
                    
                    for i in range(num_sub_agents):
                        with st.container():
                            st.write(f"**Sub-agent {i+1}:**")
                            col_name, col_desc = st.columns(2)
                            with col_name:
                                sub_name = st.text_input(f"Name", key=f"sub_name_{i}", placeholder="research_agent")
                            with col_desc:
                                sub_desc = st.text_input(f"Description", key=f"sub_desc_{i}", placeholder="Research specialist")
                            sub_url = st.text_input(f"Server URL", key=f"sub_url_{i}", placeholder="http://localhost:8011")
                            
                            if sub_name and sub_desc and sub_url:
                                sub_agents_config.append({
                                    "name": sub_name,
                                    "description": sub_desc,
                                    "server_url": sub_url
                                })
                            
                            if i < num_sub_agents - 1:  # Don't add divider after last item
                                st.divider()
                    
                    if sub_agents_config:
                        st.success(f"✅ {len(sub_agents_config)} sub-agent(s) configured")
                        with st.expander("� Deployment Order"):
                            st.markdown("""
                            **Start servers in this order:**
                            1. Start all sub-agent servers first
                            2. Wait for them to be healthy
                            3. Start the main coordinator agent
                            
                            **Example:**
                            ```bash
                            # Terminal 1: Start sub-agents
                            xagent-server --config research_agent.yaml
                            
                            # Terminal 2: Start sub-agents  
                            xagent-server --config writing_agent.yaml
                            
                            # Terminal 3: Start coordinator (this agent)
                            xagent-server --config coordinator_agent.yaml
                            ```
                            """)
                    else:
                        st.warning("⚠️ Fill in all sub-agent fields to enable multi-agent system")
            
            # Structured Output Configuration
            with st.expander("Structured Output (Optional)"):
                st.markdown("Define the expected response format using Pydantic models.")
                
                enable_structured_output = st.checkbox("Enable Structured Output")
                
                if enable_structured_output:
                    class_name = st.text_input("Class Name", value="ResponseModel")
                    
                    st.write("**Fields:**")
                    num_fields = st.number_input("Number of Fields", min_value=1, max_value=20, value=1)
                    
                    output_fields = {}
                    
                    for i in range(num_fields):
                        with st.container():
                            col_field_name, col_field_type = st.columns([1, 1])
                            
                            with col_field_name:
                                field_name = st.text_input(f"Field {i+1} Name", key=f"field_name_{i}", placeholder="title")
                            with col_field_type:
                                field_type = st.selectbox(
                                    f"Type",
                                    ["str", "int", "float", "bool", "list"],
                                    key=f"field_type_{i}"
                                )

                            # For list type, add items specification
                            list_items_type = None
                            if field_type == "list":
                                list_items_type = st.selectbox(
                                    f"List Items Type",
                                    ["str", "int", "float", "bool"],
                                    key=f"list_items_{i}",
                                    help="Type of elements in the list"
                                )
                                # st.markdown(f"*This will create: `List[{list_items_type}]`*")
                            
                            # Description field (full width)
                            field_desc = st.text_input(f"Description", key=f"field_desc_{i}", placeholder="Description of the field")
                            

                            if field_name:
                                field_config = {
                                    "type": field_type,
                                    "description": field_desc
                                }
                                
                                # Add items for list type
                                if field_type == "list" and list_items_type:
                                    field_config["items"] = {"type": list_items_type}
                                
                                output_fields[field_name] = field_config
                            
                            if i < num_fields - 1:  # Don't add divider after last field
                                st.divider()
                    
                    # Show preview of generated model
                    if output_fields:
                        st.subheader("📋 Generated Model Preview")
                        
                        # Generate Python code preview
                        python_code = f"from typing import List\nfrom pydantic import BaseModel, Field\n\nclass {class_name}(BaseModel):\n"
                        
                        for field_name, field_config in output_fields.items():
                            field_type_str = field_config["type"]
                            if field_type_str == "list" and "items" in field_config:
                                items_type = field_config["items"]["type"]
                                field_type_str = f"List[{items_type}]"
                            elif field_type_str == "str":
                                field_type_str = "str"
                            elif field_type_str == "int":
                                field_type_str = "int"
                            elif field_type_str == "float":
                                field_type_str = "float"
                            elif field_type_str == "bool":
                                field_type_str = "bool"
                            elif field_type_str == "dict":
                                field_type_str = "dict"
                            
                            python_code += f'    {field_name}: {field_type_str} = Field(description="{field_config["description"]}")\n'
                        
                        st.code(python_code, language="python")
                    else:
                        st.info("👆 Add fields above to see the generated model preview")
        
            # Session Configuration
            st.subheader("Session Settings")
            use_local_session = st.checkbox(
                "Use Local Session", 
                value=True, 
                help="If unchecked, will use Redis for session persistence (requires REDIS_URL in environment)"
            )

        with col2:
            st.subheader("📋 Configuration Preview")
            
            # Build configuration
            config = self._build_config(
                agent_name, system_prompt, model, host, port,
                enable_web_search, enable_draw_image, custom_tools,
                mcp_servers, use_local_session, toolkit_path,
                sub_agents_config, enable_structured_output,
                output_fields if 'output_fields' in locals() else {},
                class_name if 'class_name' in locals() else ""
            )
            
            # Display YAML preview
            st.code(yaml.dump(config, default_flow_style=False, allow_unicode=True), language="yaml")
            
            # Save and start buttons
            st.subheader("Actions")
            
            config_filename = st.text_input(
                "Config Filename",
                value=f"{agent_name.lower().replace(' ', '_')}_config.yaml"
            )
            
            col_save, col_start = st.columns(2)
            
            with col_save:
                if st.button("💾 Save Config", use_container_width=True):
                    config_path = self.config_dir / config_filename
                    with open(config_path, 'w', encoding='utf-8') as f:
                        yaml.dump(config, f, default_flow_style=False, allow_unicode=True)
                    st.success(f"Config saved to {config_path}")
                    time.sleep(1)
                    st.rerun()
            
            with col_start:
                if st.button("🚀 Start Server", use_container_width=True):
                    self._start_server(config, config_filename, toolkit_path)
    
    def _build_config(self, agent_name, system_prompt, model, host, port,
                     enable_web_search, enable_draw_image, custom_tools,
                     mcp_servers, use_local_session, toolkit_path,
                     sub_agents_config, enable_structured_output,
                     output_fields, class_name):
        """Build configuration dictionary."""
        config = {
            "agent": {
                "name": agent_name,
                "system_prompt": system_prompt,
                "model": model,
                "use_local_session": use_local_session
            },
            "server": {
                "host": host,
                "port": port
            }
        }
        
        # Add capabilities
        capabilities = {}
        
        # Tools
        tools = []
        if enable_web_search:
            tools.append("web_search")
        if enable_draw_image:
            tools.append("draw_image")
        
        # Custom tools
        if custom_tools.strip():
            custom_tool_list = [tool.strip() for tool in custom_tools.split('\n') if tool.strip()]
            tools.extend(custom_tool_list)
        
        if tools:
            capabilities["tools"] = tools
        
        # MCP servers
        if mcp_servers.strip():
            mcp_server_list = [server.strip() for server in mcp_servers.split('\n') if server.strip()]
            if mcp_server_list:
                capabilities["mcp_servers"] = mcp_server_list
        
        if capabilities:
            config["agent"]["capabilities"] = capabilities
        
        # Sub-agents
        if sub_agents_config:
            config["agent"]["sub_agents"] = sub_agents_config
        
        # Structured output
        if enable_structured_output and output_fields and class_name:
            config["agent"]["output_schema"] = {
                "class_name": class_name,
                "fields": output_fields
            }
        
        return config
    
    def _start_server(self, config, config_filename, toolkit_path):
        """Start the agent server."""
        try:
            # Check prerequisites
            if not self._check_prerequisites():
                return
                
            # Save config file
            config_path = self.config_dir / config_filename
            with open(config_path, 'w', encoding='utf-8') as f:
                yaml.dump(config, f, default_flow_style=False, allow_unicode=True)
            
            # Build command
            cmd = ["xagent-server", "--config", str(config_path)]
            if toolkit_path and toolkit_path.strip() and toolkit_path != "toolkit":
                cmd.extend(["--toolkit_path", toolkit_path])
            
            st.info(f"🚀 Starting server with command: {' '.join(cmd)}")
            
            # Start server process
            with st.spinner("Starting server..."):
                process = subprocess.Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    preexec_fn=os.setsid,
                    cwd=os.getcwd(),
                    env=os.environ.copy()
                )
                
                # Wait and monitor server startup
                server_url = f"http://{config['server']['host']}:{config['server']['port']}"
                max_wait_time = 30  # Maximum wait time in seconds
                check_interval = 2  # Check every 2 seconds
                waited_time = 0
                
                while waited_time < max_wait_time:
                    time.sleep(check_interval)
                    waited_time += check_interval
                    
                    # Check if process is still alive
                    if process.poll() is not None:
                        # Process has terminated
                        stdout, stderr = process.communicate()
                        stdout_str = stdout.decode() if stdout else ""
                        stderr_str = stderr.decode() if stderr else ""
                        
                        if stderr_str:
                            st.error(f"**Error output:**\n```\n{stderr_str}\n```")
                        if stdout_str:
                            st.info(f"**Standard output:**\n```\n{stdout_str}\n```")
                        
                        st.error("❌ Server process terminated unexpectedly")
                        return
                    
                    # Check if server is responding
                    if self._check_server_health(server_url):
                        # Server is up and running
                        server_id = f"{config['agent']['name']}_{config['server']['port']}"
                        self.running_servers[server_id] = {
                            "name": config['agent']['name'],
                            "url": server_url,
                            "config_file": config_filename,
                            "pid": process.pid,
                            "started_at": datetime.now().isoformat(),
                            "toolkit_path": toolkit_path
                        }
                        self._save_server_registry()
                        
                        st.success(f"✅ Server started successfully!")
                        st.info(f"🌐 Server URL: {server_url}")
                        st.info(f"🔗 Health Check: {server_url}/health")
                        st.info(f"⏱️ Startup time: {waited_time} seconds")
                        
                        return
                    
                    # Show progress
                    progress_msg = f"⏳ Waiting for server... ({waited_time}/{max_wait_time}s)"
                    st.info(progress_msg)
                
                # Timeout reached
                st.error("❌ Server startup timeout")
                
                # Try to get any output before terminating
                try:
                    stdout, stderr = process.communicate(timeout=3)
                    stdout_str = stdout.decode() if stdout else ""
                    stderr_str = stderr.decode() if stderr else ""
                    
                    if stderr_str:
                        st.error(f"**Error output:**\n```\n{stderr_str}\n```")
                    if stdout_str:
                        st.info(f"**Standard output:**\n```\n{stdout_str}\n```")
                        
                except subprocess.TimeoutExpired:
                    st.warning("Could not retrieve process output")
                
                # Terminate the process
                try:
                    process.terminate()
                    time.sleep(2)
                    if process.poll() is None:
                        process.kill()
                except:
                    pass
        
        except FileNotFoundError:
            st.error("❌ `xagent-server` command not found. Please ensure xAgent is properly installed.")
            st.info("Try running: `pip install -e .` in the xAgent directory")
        except Exception as e:
            st.error(f"❌ Error starting server: {str(e)}")
            import traceback
            st.error(f"**Traceback:**\n```\n{traceback.format_exc()}\n```")
    
    def render_server_management(self):
        """Render server management page."""
        
        # List saved configurations
        st.subheader("Saved Configurations")
        
        config_files = list(self.config_dir.glob("*.yaml"))
        config_files = [f for f in config_files if f.name != "server_registry.json"]
        
        if not config_files:
            st.info("No saved configurations found.")
            return
        
        for config_file in config_files:
            with st.expander(f"📄 {config_file.name}"):
                try:
                    with open(config_file, 'r', encoding='utf-8') as f:
                        config = yaml.safe_load(f)
                    
                    col1, col2, col3 = st.columns([2, 1, 1])
                    
                    with col1:
                        st.write(f"**Agent:** {config.get('agent', {}).get('name', 'Unknown')}")
                        st.write(f"**Model:** {config.get('agent', {}).get('model', 'Unknown')}")
                        st.write(f"**Port:** {config.get('server', {}).get('port', 'Unknown')}")
                    
                    with col2:
                        if st.button(f"🚀 Start", key=f"start_{config_file.name}"):
                            self._start_server_from_file(config_file)
                    
                    with col3:
                        if st.button(f"🗑️ Delete", key=f"delete_{config_file.name}"):
                            os.remove(config_file)
                            st.success(f"Deleted {config_file.name}")
                            time.sleep(1)
                            st.rerun()
                    
                    # Show config content
                    with st.expander("View Configuration"):
                        st.code(yaml.dump(config, default_flow_style=False, allow_unicode=True), language="yaml")
                
                except Exception as e:
                    st.error(f"Error reading config file: {str(e)}")
    
    def _start_server_from_file(self, config_file):
        """Start server from saved configuration file."""
        try:
            with open(config_file, 'r', encoding='utf-8') as f:
                config = yaml.safe_load(f)
            
            toolkit_path = "toolkit"  # Default toolkit path
            self._start_server(config, config_file.name, toolkit_path)
            time.sleep(1)
            st.rerun()
        
        except Exception as e:
            st.error(f"Error starting server: {str(e)}")
    
    def render_running_servers(self):
        """Render running servers management page."""
        st.subheader("🚀 Running Servers")
        
        # Clean up dead servers first
        self._cleanup_dead_servers()
        self._cleanup_dead_webchats()
        
        if not self.running_servers:
            st.info("No running servers found.")
            return
        
        # Refresh button
        if st.button("🔄 Refresh", use_container_width=False):
            st.rerun()
        
        st.divider()
        
        for server_id, server_info in self.running_servers.items():
            with st.container():
                # Check server health
                is_healthy = self._check_server_health(server_info['url'])
                status_icon = "🟢" if is_healthy else "🔴"
                status_text = "Healthy" if is_healthy else "Unhealthy"
                
                # Check if this server has a running web chat
                webchat_info = None
                for wc_id, wc_info in self.running_webchats.items():
                    if wc_info.get('agent_server') == server_info['url']:
                        webchat_info = wc_info
                        break
                
                col1, col2, col3, col4, col5 = st.columns([2, 1, 1, 1, 1])
                
                with col1:
                    st.write(f"{status_icon} **{server_info['name']}**")
                    st.write(f"URL: {server_info['url']}")
                    st.write(f"Started: {server_info.get('started_at', 'Unknown')}")
                    
                    # Show web chat status if exists
                    if webchat_info:
                        webchat_healthy = self._check_webchat_health(webchat_info['url'])
                        webchat_icon = "🟢" if webchat_healthy else "🔴"
                        st.write(f"{webchat_icon} **Web Chat:** {webchat_info['url']}")
                
                with col2:
                    st.write(f"**Status:** {status_text}")
                    st.write(f"**PID:** {server_info.get('pid', 'Unknown')}")
                    if webchat_info:
                        st.write(f"**Chat PID:** {webchat_info.get('pid', 'Unknown')}")
                
                with col3:
                    if st.button("🌐 Open", key=f"open_{server_id}"):
                        st.markdown(f"[Open Server]({server_info['url']}/health)")
                
                with col4:
                    if webchat_info:
                        # Show close web chat button if web chat is running
                        if st.button("❌ Close Chat", key=f"closechat_{server_id}"):
                            self._stop_web_chat(webchat_info)
                            time.sleep(1)
                            st.rerun()
                    else:
                        # Show start web chat button if no web chat is running
                        if st.button("💬 Web Chat", key=f"webchat_{server_id}"):
                            if is_healthy:
                                self._start_web_chat(server_info['url'])
                                time.sleep(1)
                                st.rerun()
                            else:
                                st.error("❌ Server is not healthy. Please check server status first.")
                
                with col5:
                    if st.button("🛑 Stop", key=f"stop_{server_id}"):
                        # Also stop web chat if running
                        if webchat_info:
                            self._stop_web_chat(webchat_info)
                        self._stop_server(server_id, server_info)
                        time.sleep(1)
                        st.rerun()
                
                # Server details
                with st.expander(f"Details - {server_info['name']}"):
                    st.json(server_info)
                
                st.divider()
    
    def _stop_server(self, server_id, server_info):
        """Stop a running server."""
        try:
            pid = server_info.get('pid')
            if pid:
                try:
                    # Try graceful termination first
                    os.kill(pid, signal.SIGTERM)
                    time.sleep(2)
                    
                    # Force kill if still running
                    if psutil.pid_exists(pid):
                        os.kill(pid, signal.SIGKILL)
                    
                    st.success(f"✅ Server {server_info['name']} stopped successfully")
                
                except ProcessLookupError:
                    st.info(f"Server {server_info['name']} was already stopped")
                except Exception as e:
                    st.error(f"Error stopping server: {str(e)}")
            
            # Remove from registry
            del self.running_servers[server_id]
            self._save_server_registry()
        
        except Exception as e:
            st.error(f"Failed to stop server: {str(e)}")
    
    def _start_web_chat(self, agent_url):
        """Start web chat interface for the agent server."""
        try:
            # Convert server URL for web chat command
            # Handle different host formats
            if "0.0.0.0" in agent_url:
                # Replace 0.0.0.0 with localhost for web chat
                chat_agent_url = agent_url.replace("0.0.0.0", "localhost")
            else:
                chat_agent_url = agent_url
            
            # Find available port
            try:
                available_port = self._find_available_port(8501)
                st.info(f"🔍 Found available port: {available_port}")
            except Exception as e:
                st.error(f"❌ Could not find available port: {str(e)}")
                return
            
            # Build web chat command
            cmd = [
                "xagent-web",
                "--host", "0.0.0.0",
                "--port", str(available_port), 
                "--agent-server", chat_agent_url
            ]
            
            st.info(f"🚀 Starting Web Chat with command: {' '.join(cmd)}")
            
            # Start web chat process in background
            with st.spinner("Starting Web Chat..."):
                process = subprocess.Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    preexec_fn=os.setsid,
                    cwd=os.getcwd(),
                    env=os.environ.copy()
                )
                
                # Wait a moment for startup
                time.sleep(5)
                
                # Check if process is still alive
                if process.poll() is not None:
                    # Process has terminated
                    stdout, stderr = process.communicate()
                    stdout_str = stdout.decode() if stdout else ""
                    stderr_str = stderr.decode() if stderr else ""
                    
                    if stderr_str:
                        st.error(f"**Error output:**\n```\n{stderr_str}\n```")
                    if stdout_str:
                        st.info(f"**Standard output:**\n```\n{stdout_str}\n```")
                    
                    st.error("❌ Web Chat process terminated unexpectedly")
                    return
                
                # Check if web chat is responding
                web_chat_url = f"http://localhost:{available_port}"
                max_wait_time = 15  # Maximum wait time in seconds
                check_interval = 2  # Check every 2 seconds
                waited_time = 0
                
                while waited_time < max_wait_time:
                    if self._check_webchat_health(web_chat_url):
                        # Web Chat is up and running
                        webchat_id = f"webchat_{agent_url.split(':')[-1]}_{available_port}"
                        self.running_webchats[webchat_id] = {
                            "agent_server": agent_url,
                            "url": web_chat_url,
                            "port": available_port,
                            "pid": process.pid,
                            "started_at": datetime.now().isoformat()
                        }
                        self._save_webchat_registry()
                        
                        st.success(f"✅ Web Chat started successfully!")
                        st.info(f"💬 Web Chat URL: {web_chat_url}")
                        st.info(f"🔗 Agent Server: {chat_agent_url}")
                        st.info(f"⏱️ Startup time: {waited_time + 5} seconds")
                        
                        # Add a clickable link
                        st.markdown(f"👉 [Open Web Chat Interface]({web_chat_url})")
                        
                        return
                    
                    time.sleep(check_interval)
                    waited_time += check_interval
                
                # Timeout reached
                st.error("❌ Web Chat startup timeout")
                
                # Terminate the process
                try:
                    process.terminate()
                    time.sleep(2)
                    if process.poll() is None:
                        process.kill()
                except:
                    pass
                
        except FileNotFoundError:
            st.error("❌ `xagent-web` command not found. Please ensure xAgent is properly installed.")
            st.info("Try running: `pip install -e .` in the xAgent directory")
        except Exception as e:
            st.error(f"❌ Error starting Web Chat: {str(e)}")
            import traceback
            st.error(f"**Traceback:**\n```\n{traceback.format_exc()}\n```")
    
    def _stop_web_chat(self, webchat_info):
        """Stop a running web chat."""
        try:
            pid = webchat_info.get('pid')
            port = webchat_info.get('port')
            
            if pid:
                try:
                    # First, try to terminate the entire process group
                    try:
                        # Get the process group ID and terminate the whole group
                        os.killpg(os.getpgid(pid), signal.SIGTERM)
                        st.info(f"🔄 Sending SIGTERM to process group {pid}...")
                        time.sleep(3)
                        
                        # Check if process still exists
                        if psutil.pid_exists(pid):
                            st.warning(f"⚠️ Process {pid} still running, force killing...")
                            os.killpg(os.getpgid(pid), signal.SIGKILL)
                            time.sleep(2)
                    except ProcessLookupError:
                        # Process group doesn't exist, try individual process
                        os.kill(pid, signal.SIGTERM)
                        time.sleep(3)
                        
                        if psutil.pid_exists(pid):
                            os.kill(pid, signal.SIGKILL)
                            time.sleep(2)
                    
                    # Double check: kill any remaining processes using the port
                    if port:
                        self._kill_processes_on_port(port)
                    
                    # Verify the process is really dead
                    if psutil.pid_exists(pid):
                        st.error(f"❌ Failed to terminate process {pid}")
                    else:
                        st.success(f"✅ Web Chat on port {port} stopped successfully")
                
                except ProcessLookupError:
                    st.info(f"Web Chat on port {port} was already stopped")
                except Exception as e:
                    st.error(f"Error stopping Web Chat: {str(e)}")
                    # Try to kill processes on port as fallback
                    if port:
                        self._kill_processes_on_port(port)
            
            # Remove from registry
            webchat_id_to_remove = None
            for wc_id, wc_info in self.running_webchats.items():
                if wc_info.get('pid') == webchat_info.get('pid'):
                    webchat_id_to_remove = wc_id
                    break
            
            if webchat_id_to_remove:
                del self.running_webchats[webchat_id_to_remove]
                self._save_webchat_registry()
                
            # Final verification: check if port is really free
            if port:
                time.sleep(1)  # Wait a moment for port to be released
                if self._is_port_in_use(port):
                    st.warning(f"⚠️ Port {port} may still be in use")
                else:
                    st.info(f"✅ Port {port} is now available")
        
        except Exception as e:
            st.error(f"Failed to stop Web Chat: {str(e)}")
    
    def _kill_processes_on_port(self, port):
        """Kill all processes using the specified port."""
        try:
            import subprocess
            
            # Find processes using the port
            result = subprocess.run(
                ['lsof', '-ti', f':{port}'],
                capture_output=True,
                text=True
            )
            
            if result.returncode == 0 and result.stdout.strip():
                pids = result.stdout.strip().split('\n')
                for pid_str in pids:
                    try:
                        pid = int(pid_str.strip())
                        os.kill(pid, signal.SIGKILL)
                        st.info(f"🔫 Force killed process {pid} using port {port}")
                    except (ValueError, ProcessLookupError):
                        continue
        except FileNotFoundError:
            # lsof not available, try alternative method
            try:
                result = subprocess.run(
                    ['netstat', '-tulpn'],
                    capture_output=True,
                    text=True
                )
                # Parse netstat output to find PIDs using the port
                # This is a more complex parsing, but provides fallback
                lines = result.stdout.split('\n')
                for line in lines:
                    if f':{port}' in line and 'LISTEN' in line:
                        parts = line.split()
                        if len(parts) > 6 and '/' in parts[-1]:
                            try:
                                pid = int(parts[-1].split('/')[0])
                                os.kill(pid, signal.SIGKILL)
                                st.info(f"🔫 Force killed process {pid} using port {port}")
                            except (ValueError, ProcessLookupError):
                                continue
            except:
                st.warning(f"⚠️ Could not find processes using port {port}")
        except Exception as e:
            st.warning(f"⚠️ Error killing processes on port {port}: {str(e)}")
    
    def _is_port_in_use(self, port):
        """Check if a port is currently in use."""
        import socket
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(('localhost', port))
                return False
        except OSError:
            return True


def main():
    """Main function to run the Streamlit app."""
    config_ui = AgentConfigUI()
    config_ui.render_main_page()


if __name__ == "__main__":
    main()



import os
import yaml
import argparse
import asyncio
import uuid
import logging
from typing import Optional, Dict, Any
from dotenv import load_dotenv

load_dotenv(override=True)

from ..core.base import BaseAgentRunner
from ..core.session import Session


class CLIAgent(BaseAgentRunner):
    """CLI Agent for xAgent."""
    
    def __init__(self, config_path: Optional[str] = None, toolkit_path: Optional[str] = None, verbose: bool = False):
        """
        Initialize CLIAgent.
        
        Args:
            config_path: Path to configuration file (if None, uses default configuration)
            toolkit_path: Path to toolkit directory (if None, no additional tools will be loaded)
            verbose: Enable verbose logging output
        """
        # Configure logging based on verbose setting
        self.verbose = verbose
        if not verbose:
            # Suppress most logging except critical errors
            logging.getLogger().setLevel(logging.CRITICAL)
            logging.getLogger("xagent").setLevel(logging.CRITICAL)
        else:
            # Enable verbose logging
            logging.getLogger().setLevel(logging.INFO)
            logging.getLogger("xagent").setLevel(logging.INFO)
        
        # Initialize the base agent runner
        super().__init__(config_path, toolkit_path)
        
        # Store config_path for CLI-specific functionality
        self.config_path = config_path if config_path and os.path.isfile(config_path) else None
        
    async def chat_interactive(self, user_id: str = None, session_id: str = None, stream: bool = None):
        """
        Start an interactive chat session.
        
        Args:
            user_id: User ID for the session
            session_id: Session ID for the chat
            stream: Enable streaming response (default: True, but False when verbose mode is enabled)
        """
        # If stream is not explicitly set, determine based on verbose mode
        if stream is None:
            # When verbose mode is enabled, default to non-streaming for better log readability
            stream = not (logging.getLogger().level <= logging.INFO)
        
        # Check if verbose mode is enabled by checking log level
        verbose_mode = logging.getLogger().level <= logging.INFO
        # Generate default IDs if not provided
        user_id = user_id or f"cli_user_{uuid.uuid4().hex[:8]}"
        session_id = session_id or f"cli_session_{uuid.uuid4().hex[:8]}"
        
        session = Session(
            user_id=user_id,
            session_id=session_id,
            message_db=self.message_db
        )
        
        print(f"🤖 Welcome to xAgent CLI!")
        config_msg = f"Loading agent configuration from {self.config_path}" if self.config_path else "Using default configuration"
        print(config_msg)
        print(f"Agent: {self.agent.name}")
        print(f"Model: {self.agent.model}")
        print(f"Tools: {len(self.agent.tools)} loaded")
        print(f"Session: {session_id}")
        print(f"Verbose mode: {'Enabled' if verbose_mode else 'Disabled'}")
        print(f"Streaming: {'Enabled' if stream else 'Disabled'}")
        if verbose_mode and stream:
            print("ℹ️  Note: Verbose mode is enabled. Consider using 'stream off' for better log readability.")
        print("Type 'exit', 'quit', or 'bye' to end the session.")
        print("Type 'clear' to clear the session history.")
        print("Type 'stream on/off' to toggle streaming mode.")
        print("Type 'help' for available commands.")
        print("-" * 50)
        
        while True:
            try:
                # Get user input
                user_input = input("\n👤 You: ").strip()
                
                # Handle special commands
                if user_input.lower() in ['exit', 'quit', 'bye']:
                    print("👋 Goodbye!")
                    break
                elif user_input.lower() == 'clear':
                    await session.clear_session()
                    print("🧹 Session history cleared.")
                    continue
                elif user_input.lower().startswith('stream '):
                    # Handle stream toggle command
                    stream_cmd = user_input.lower().split()
                    if len(stream_cmd) == 2:
                        if stream_cmd[1] == 'on':
                            stream = True
                            print("🌊 Streaming mode enabled.")
                        elif stream_cmd[1] == 'off':
                            stream = False
                            print("📄 Streaming mode disabled.")
                        else:
                            print("⚠️  Usage: stream on/off")
                    else:
                        print("⚠️  Usage: stream on/off")
                    continue
                elif user_input.lower() == 'help':
                    self._show_help()
                    continue
                elif not user_input:
                    continue
                
                # Process the message
                if stream:
                    # Handle streaming response
                    response_generator = await self.agent(
                        user_message=user_input,
                        session=session,
                        stream=True
                    )
                    
                    # Check if response is a generator (streaming) or a string
                    if hasattr(response_generator, '__aiter__'):
                        print("🤖 Agent: ", end="", flush=True)
                        async for chunk in response_generator:
                            if chunk:
                                print(chunk, end="", flush=True)
                        print()  # Add newline after streaming is complete
                    else:
                        # Fallback for non-streaming response
                        print("🤖 Agent: " + str(response_generator))
                else:
                    # Handle non-streaming response
                    response = await self.agent(
                        user_message=user_input,
                        session=session,
                        stream=False
                    )
                    print("🤖 Agent: " + str(response))
                
            except KeyboardInterrupt:
                print("\n\n👋 Session interrupted. Goodbye!")
                break
            except Exception as e:
                print(f"\n❌ Error: {e}")
    
    async def chat_single(self, message: str, user_id: str = None, session_id: str = None):
        """
        Process a single message and return the response.
        
        Args:
            message: The message to process
            user_id: User ID for the session
            session_id: Session ID for the chat
            
        Returns:
            Agent response string
        """
        # Generate default IDs if not provided
        user_id = user_id or f"cli_user_{uuid.uuid4().hex[:8]}"
        session_id = session_id or f"cli_session_{uuid.uuid4().hex[:8]}"
        
        session = Session(
            user_id=user_id,
            session_id=session_id,
            message_db=self.message_db
        )
        
        response = await self.agent(
            user_message=message,
            session=session,
            stream=False
        )
        
        return response
    
    def _show_help(self):
        """Show help information."""
        print("\n📋 Available commands:")
        print("  exit, quit, bye  - Exit the chat session")
        print("  clear           - Clear session history")
        print("  stream on/off   - Toggle streaming mode")
        print("  help            - Show this help message")
        print("\n🔧 Available tools:")
        for tool_name in self.agent.tools.keys():
            print(f"  - {tool_name}")
        if self.agent.mcp_tools:
            print("\n🌐 MCP tools:")
            for tool_name in self.agent.mcp_tools.keys():
                print(f"  - {tool_name}")
    
    def create_default_config(self, config_path: str = "config/agent.yaml"):
        """
        Create a default configuration file.
        
        Args:
            config_path: Path where to create the config file
        """
        # Create directory if it doesn't exist
        config_dir = os.path.dirname(config_path)
        if config_dir and not os.path.exists(config_dir):
            os.makedirs(config_dir)
        
        default_config = self._get_default_config()
        
        with open(config_path, 'w', encoding='utf-8') as f:
            yaml.dump(default_config, f, default_flow_style=False, allow_unicode=True)
        
        print(f"✅ Default configuration created at: {config_path}")
        print("You can edit this file to customize your agent settings.")


def create_default_config_file(config_path: str = "config/agent.yaml"):
    """
    Create a default configuration file.
    
    Args:
        config_path: Path where to create the config file
    """
    # Create directory if it doesn't exist
    config_dir = os.path.dirname(config_path)
    if config_dir and not os.path.exists(config_dir):
        os.makedirs(config_dir)
    
    # Use the default configuration from BaseAgentRunner
    from ..core.base import BaseAgentRunner
    dummy_runner = BaseAgentRunner()
    default_config = dummy_runner._get_default_config()
    
    with open(config_path, 'w', encoding='utf-8') as f:
        yaml.dump(default_config, f, default_flow_style=False, allow_unicode=True)
    
    print(f"✅ Default configuration created at: {config_path}")
    print("You can edit this file to customize your agent settings.")


def main():
    """Main entry point for xagent-cli command."""
    parser = argparse.ArgumentParser(description="xAgent CLI")
    
    # Subcommands
    subparsers = parser.add_subparsers(dest="command", help="Available commands")
    
    # Interactive chat command
    chat_parser = subparsers.add_parser("chat", help="Start interactive chat session")
    chat_parser.add_argument("--config", default=None, help="Config file path (if not specified, uses default configuration)")
    chat_parser.add_argument("--toolkit_path", default=None, help="Toolkit directory path (if not specified, no additional tools will be loaded)")
    chat_parser.add_argument("--user_id", help="User ID for the session")
    chat_parser.add_argument("--session_id", help="Session ID for the chat")
    chat_parser.add_argument("--verbose", "-v", action="store_true", help="Enable verbose logging")
    chat_parser.add_argument("--no-stream", action="store_true", help="Disable streaming response (default: streaming enabled)")
    
    # Single message command
    single_parser = subparsers.add_parser("ask", help="Ask a single question")
    single_parser.add_argument("message", help="The message to send to the agent")
    single_parser.add_argument("--config", default=None, help="Config file path (if not specified, uses default configuration)")
    single_parser.add_argument("--toolkit_path", default=None, help="Toolkit directory path (if not specified, no additional tools will be loaded)")
    single_parser.add_argument("--user_id", help="User ID for the session")
    single_parser.add_argument("--session_id", help="Session ID for the chat")
    single_parser.add_argument("--verbose", "-v", action="store_true", help="Enable verbose logging")
    
    # Init command to create default config
    init_parser = subparsers.add_parser("init", help="Create default configuration file")
    init_parser.add_argument("--config", default="config/agent.yaml", help="Config file path to create")
    
    # Parse arguments
    args = parser.parse_args()
    
    # If no command specified, default to interactive chat
    if not args.command:
        args.command = "chat"
        args.config = None  # Use default configuration when no command is specified
        args.toolkit_path = None  # No toolkit by default
        args.user_id = None
        args.session_id = None
        args.verbose = False
        args.no_stream = False
    
    try:
        if args.command == "init":
            create_default_config_file(args.config)
            return
            
        cli_agent = CLIAgent(
            config_path=args.config, 
            toolkit_path=args.toolkit_path,
            verbose=getattr(args, 'verbose', False)
        )
        
        if args.command == "chat":
            # Determine stream setting - if --no-stream is specified, use False
            # Otherwise, let chat_interactive decide based on verbose mode
            stream_setting = None if not getattr(args, 'no_stream', False) else False
            asyncio.run(cli_agent.chat_interactive(
                user_id=args.user_id,
                session_id=args.session_id,
                stream=stream_setting
            ))
        elif args.command == "ask":
            response = asyncio.run(cli_agent.chat_single(
                message=args.message,
                user_id=args.user_id,
                session_id=args.session_id
            ))
            print(response)
        else:
            parser.print_help()
            
    except Exception as e:
        print(f"Failed to start CLI: {e}")
        raise


if __name__ == "__main__":
    main()
